import jax.numpy as jnp
from jax.scipy.linalg import expm
from tinygp.helpers import JAXArray


def block_view(A, b):
    Nb, Mb = A.shape
    assert Nb % b == 0 and Mb % b == 0
    N = Nb // b
    M = Mb // b
    return A.reshape(N, b, M, b).transpose(0, 2, 1, 3)


def Q_from_VanLoan(F: JAXArray, L: JAXArray, Qc: JAXArray, dt: JAXArray) -> JAXArray:
    """
    Van Loan method to compute Q = ∫0^dt exp(F (dt-s)) L Qc L^T exp(F^T (dt-s)) ds

    Parameters:
        F: StateSpaceModel.design_matrix
        L: StateSpaceModel.noise_effect_matrix
        Qc: StateSpaceModel.process_noise
        dt: time step between measurements (dt = X2 - X1)

    Returns:
        Q: Process noise covariance matrix over time step dt

    See Van Loan (1978) "Computing Integrals Involving the Matrix Exponential"
    PDF at https://www.olemartin.no/artikler/vanloan.pdf
    https://ecommons.cornell.edu/items/cba38b2e-6ad4-45e6-8109-0a019fe5114c
    """
    QL = L @ Qc @ L.T
    b = len(F)  # block size
    Z = jnp.zeros_like(F)
    C = jnp.block([[-F, QL], [Z, F.T]])
    VanLoanBlock = expm(C * dt)
    G2 = VanLoanBlock[:b, b:]
    F3 = VanLoanBlock[b:, b:]
    return F3.T @ G2


def Phibar_from_VanLoan(F: JAXArray, dt: JAXArray) -> JAXArray:
    """
    Van Loan method to compute Phibar = ∫0^dt exp(F s) ds

    Parameters:
        F: StateSpaceModel.design_matrix
        dt: time step between measurements (dt = X2 - X1)

    Returns:
        Phibar: The integrated transition matrix over time step dt

    See Van Loan (1978) "Computing Integrals Involving the Matrix Exponential"
    PDF at https://www.olemartin.no/artikler/vanloan.pdf
    https://ecommons.cornell.edu/items/cba38b2e-6ad4-45e6-8109-0a019fe5114c
    """
    b = len(F)  # block size
    Z = jnp.zeros((b, b))
    I = jnp.eye(b)
    C = jnp.block([[F, I], [Z, Z]])
    VanLoanBlock = expm(C * dt)
    G3 = VanLoanBlock[:b, b:]
    return G3


def VanLoan(
    F: JAXArray, L: JAXArray, Qc: JAXArray, dt: JAXArray
) -> tuple[JAXArray, JAXArray]:
    """
    Constructs the full matrix C and returns its matrix exponential

    Parameters:
        F: StateSpaceModel.design_matrix
        L: StateSpaceModel.noise_effect_matrix
        Qc: StateSpaceModel.process_noise
        dt: time step between measurements (dt = X2 - X1)

    Returns:
        dict : Dictionary of the submatrices of the Van Loan exponential, from which
                                various integrals (e.g. Q, Phibar) can be computed.

    See Van Loan (1978) "Computing Integrals Involving the Matrix Exponential"
    PDF at https://www.olemartin.no/artikler/vanloan.pdf
    https://ecommons.cornell.edu/items/cba38b2e-6ad4-45e6-8109-0a019fe5114c
    """
    QL = L @ Qc @ L.T
    b = len(F)  # block size
    I = jnp.eye(b)
    Z = jnp.zeros_like(F)
    C = jnp.block(
        [
            [-F, I, Z, Z],
            [Z, -F, QL, Z],
            [Z, Z, F.T, I],
            [Z, Z, Z, Z],
        ]
    )
    VanLoanBlock = block_view(expm(C * dt), b)

    F1 = VanLoanBlock[0, 0]
    G1 = VanLoanBlock[0, 1]
    H1 = VanLoanBlock[0, 2]
    K1 = VanLoanBlock[0, 3]
    F2 = VanLoanBlock[1, 1]
    G2 = VanLoanBlock[1, 2]
    H2 = VanLoanBlock[1, 3]
    F3 = VanLoanBlock[2, 2]
    G3 = VanLoanBlock[2, 3]
    F4 = VanLoanBlock[3, 3]

    return {
        "F1": F1,
        "F2": F2,
        "F3": F3,
        "F4": F4,
        "G1": G1,
        "G2": G2,
        "G3": G3,
        "H1": H1,
        "H2": H2,
        "K1": K1,
    }



"""
The primary model building interface in ``smolgp`` is via "kernels", which are
typically constructed as sums and products of objects defined in this
subpackage, or by subclassing :class:`Kernel` as discussed in the :ref:`kernels`
tutorial. The kernels implemented here are extensions of those defined in
``tinygp.kernels.quasisep`` to be fully-compatible with the state space solvers.
The key differences are that `smolgp` kernels:
1. Include the `noise_effect_matrix` and `process_noise` matrix
2. Treat the observation model as a row vector that can be selected on/off via the kernel's name
3. The transition matrix is in its usual form (compared to transposed forms in tinygp.kernels.quasisep)

For modeling integrated measurements, use the kernels in ``smolgp.kernels.integrated``.

Sums and Products of ``smolgp`` kernels are also ``smolgp`` kernels.
Their conditioned and predictive distributions can be decomposed per-component
using GaussianProcess.component_means and predict_component_means, respectively.

For mixed kernels containing some integrated and some instantaneous components,
the integrated solver will be used to handle the entire kernel.
"""

# Model class and utilities
from smolgp.kernels.base import (
    StateSpaceModel as StateSpaceModel,
    Sum as Sum,
    Product as Product,
    Wrapper as Wrapper,
    Scale as Scale,
)

# Specific kernels
from smolgp.kernels.base import (
    Constant as Constant,
    SHO as SHO,
    Exp as Exp,
    Matern32 as Matern32,
    Matern52 as Matern52,
    Cosine as Cosine,
    ExpSineSquared as ExpSineSquared,
)

# Integrated kernels
from smolgp.kernels.integrated import (
    IntegratedSHO as IntegratedSHO,
    IntegratedExp as IntegratedExp,
    IntegratedMatern32 as IntegratedMatern32,
    IntegratedMatern52 as IntegratedMatern52,
    IntegratedCosine as IntegratedCosine,
)

# Dense kernels (for testing)
import smolgp.kernels.dense



import jax
import jax.numpy as jnp
import tinygp

key = jax.random.PRNGKey(0)


def allclose(name, residuals, tol, atol=1e-14):
    """
    Check all residuals are < tol
    if they are, but aren't < atol, print a warning
    """
    maxres = jnp.max(jnp.abs(residuals))
    assert maxres < tol, (
        f"{name} did not agree to within desired tolerance."
        f" Maximum absolute deviation is {maxres:.3e} "
    )
    if maxres < atol:
        print(f"    ...{name}: agrees exactly (<{maxres:.0e})")
    else:
        print(f"    ...{name}: agrees (WARNING: only to < {maxres:.1e})")


def format_bytes(n):
    if n == 0:
        return "0 B"
    units = ["B", "KB", "MB", "GB", "TB", "PB"]
    size = float(n)
    idx = 0
    while size >= 1024 and idx < len(units) - 1:
        size /= 1024
        idx += 1
    return f"{size:.1f} {units[idx]}"


def generate_data(N, kernel, yerr=0.3, tmin=0, tmax=86400):
    t_train = jnp.linspace(tmin, tmax, N)
    true_gp = tinygp.GaussianProcess(kernel, t_train)
    y_true = true_gp.sample(key=key)
    y_train = y_true + yerr * jax.random.normal(key, shape=(N,))
    return t_train, y_train


def generate_integrated_data(N, kernel, texp=180, yerr=0.3, readout=40):
    # Generate true GP over baseline
    cadence = texp + readout
    tmin = 0
    tmax = N * cadence
    buffer = 0.1 * (tmax - tmin)
    t = jnp.arange(tmin - buffer, tmax + buffer, 1)
    true_gp = tinygp.GaussianProcess(kernel, t)
    y = true_gp.sample(key=key)

    # Generate synthetic observations
    @jax.jit
    def make_exposure(tmid, texp):
        t_in_exp = jnp.linspace(tmid - texp / 2, tmid + texp / 2, 50)
        # quickly slice region of interest
        idx = jnp.searchsorted(t, t_in_exp, side="right")
        idx = jnp.clip(idx, 0, t.size - 2)
        # just interpolate that region
        y_in_exp = jnp.interp(t_in_exp, t[idx], y[idx])
        return jnp.mean(y_in_exp)

    texp_train = jnp.full(N, texp)  # constant exposure time
    t_train = jnp.arange(tmin, tmax, cadence)
    y_true = jax.vmap(make_exposure)(t_train, texp_train)
    y_train = y_true + yerr * jax.random.normal(key, shape=(len(t_train),))
    return t_train, y_train


def get_data(true_kernel, N, yerr=0.3, exposure_quantities=None, save=True):
    # Generate data of length N
    if exposure_quantities:
        texp, readout = exposure_quantities
        t_train, y_train = generate_integrated_data(
            N, true_kernel, texp=texp, readout=readout, yerr=yerr
        )
        texp_train = jnp.full_like(t_train, texp)
        yerr_train = jnp.full_like(t_train, yerr)
        instid = jnp.full_like(t_train, 0)
        data = jnp.array([t_train, y_train, yerr_train, texp_train, instid])
        savename = f"data/{N}_int.npz"
    else:
        t_train, y_train = generate_data(N, true_kernel, yerr=yerr)
        yerr_train = jnp.full_like(t_train, yerr)
        data = jnp.array([t_train, y_train, yerr_train])
        savename = f"data/{N}.npz"
    if save:
        jnp.savez(savename, data)
    return data



# TODO:
# 1. copy base.py kernel object style
# 2. add integrated_transition_matrix and integrated_process_noise
# 3. add attribute/property for num_insts
# 4. define each of the usual matrix components to be the augmented version
#    e.g. stationary_covariance --> BlockDiag(sho.stationary_covariance, identity)

# in the solver, user will have passed t, texp, instid, and y
# from there, stateid will get auto-created according to t and texp

"""
These kernels are compatible with :class:`smolgp.solvers.integrated.IntegratedStateSpaceSolver`,
which uses Bayesian filtering and smoothing algorithms to perform scalable GP
inference. (see :ref:`api-solvers-statespace` for more technical details).
On GPU, a performance boost may be observed for large datasets by using the
:class:`smolgp.solvers.parallel.ParallelStateSpaceSolver` class.

Like the quasisep kernels, these methods are experimental, so you may find
the documentation patchy in places. You are encouraged to `open issues or
pull requests <https://github.com/rrubenza/smolgp/issues>`_ as you find gaps.
"""

from __future__ import annotations

__all__ = [
    "IntegratedSHO",
    "IntegratedExp",
    "IntegratedMatern32",
    "IntegratedMatern52",
    "IntegratedCosine",
]

import equinox as eqx
import jax
import jax.numpy as jnp

from tinygp.helpers import JAXArray

import smolgp.kernels
from smolgp.kernels import StateSpaceModel
from smolgp.helpers import Phibar_from_VanLoan


class IntegratedStateSpaceModel(StateSpaceModel):
    """
    A generic class that augments a :class:`StateSpaceModel`
    object (which has state `x`) with an integral state `z`,
    to model the joint state `X = [x; z]`.

    The coordinates for an integrated model should be a tuple of
        X = (t, delta, instid),
    where `t` is the usual coordinate (e.g. time) at the measurements (midpoints),
    `delta` is the integration range (e.g. exposure time) for each measurement,
    and `instid` is an index encoding which instrument the measurement corresponds to.
    """

    base_model: StateSpaceModel  # the base (non-integrated) SSM
    num_insts: int = eqx.field(static=True)  # number of integral states

    @property
    def d(self) -> int:
        """The dimension of the base (non-integrated) state space model"""
        return self.base_model.dimension

    @property
    def dimension(self) -> int:
        """The dimension of the augmented state space model"""
        return self.d + self.num_insts

    def coord_to_sortable(self, X: JAXArray) -> JAXArray:
        """
        A helper function used to convert coordinates to sortable 1-D values

        If X is a tuple, e.g. of (time, delta, instid), this assumes the first coordinate is the sortable one
        """
        if isinstance(X, tuple):
            return X[0]
        else:
            return X

    def design_matrix(self) -> JAXArray:
        """The augmented design (also called the feedback) matrix for the process, $F$"""
        F = self.base_model.design_matrix()
        F_aug = jnp.zeros((self.dimension, self.dimension))
        F_aug = F_aug.at[: self.d, : self.d].set(F)
        for i in range(self.num_insts):
            F_aug = F_aug.at[self.d + i, 0].set(1.0)
        return F_aug

    def stationary_covariance(self) -> JAXArray:
        """The augmented stationary covariance of the process, Pinf"""
        Pinf = self.base_model.stationary_covariance()
        Pinf_aug = jnp.diag(jnp.ones(self.dimension)).at[: self.d, : self.d].set(Pinf)
        return Pinf_aug

    def observation_matrix(self, X: JAXArray) -> JAXArray:
        """The augmented observation model for the process, $H$"""

        ## TODO: make sure this works for multivariate data, e.g. like:
        # H_base = self.base_model.observation_model(t)
        # H_z = H_base/delta # observe the average value over exposure
        # H_aug = jnp.zeros((H_base.shape[0], self.dimension))
        # H_aug = jax.lax.dynamic_update_slice(H_aug, H_z, (self.d*(1+instid),))
        ## Below is hardcoded for 1-D data

        def H_integral(t: JAXArray, delta: JAXArray, instid: int) -> JAXArray:
            """Observation model for integral state"""
            H_z = jnp.array([1.0 / delta])
            H_aug = jnp.zeros(self.dimension)
            H_aug = jax.lax.dynamic_update_slice(H_aug, H_z, (self.d + instid,))
            return H_aug

        def H_latent(t: JAXArray, instid: int) -> JAXArray:
            """Observation model for latent (non-integral) state"""
            # H_x = self.base_model.observation_model(X) # TODO: use this to get the shapes right
            H_x = jnp.zeros(self.d).at[0].set(1)  # hardcoded 1-D version for now
            H_aug = jnp.zeros(self.dimension)
            H_aug = jax.lax.dynamic_update_slice(H_aug, H_x, (0,))
            return H_aug

        if isinstance(X, tuple) or isinstance(X, list):
            # Observing integral state (z) with exposure time (delta)
            t, delta, instid = X
            H_aug = jax.lax.cond(
                delta > 0,
                lambda _: H_integral(t, delta, instid),
                lambda _: H_latent(t, instid),
                operand=None,
            )
        else:
            # default to latent state if no exposure time provided
            H_aug = H_latent(X, instid=0)

        return jnp.array([H_aug])

    def noise(self) -> JAXArray:
        """The spectral density of the white noise process, $Q_c$"""
        return self.base_model.noise()

    def noise_effect_matrix(self) -> JAXArray:
        """The augmented noise effect matrix, $L$"""
        L = self.base_model.noise_effect_matrix()
        L_aug = jnp.vstack([L] + [0.0] * self.num_insts)
        return L_aug

    def integrated_transition_matrix(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        """
        The integrated transition matrix between two states at coordinates X1 and X2, $A_k$

        By default uses the Van Loan method to compute Phibar = ∫0^dt exp(F s) ds

        Overload this method if you wish to define the integrated transition matrix analytically.
        """
        F = self.base_model.design_matrix()
        t1 = self.coord_to_sortable(X1)
        t2 = self.coord_to_sortable(X2)
        dt = t2 - t1
        return Phibar_from_VanLoan(F, dt)

    def transition_matrix(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        """
        The augmented transition matrix between two states at coordinates X1 and X2, $A_k$
        """
        t1 = self.coord_to_sortable(X1)
        t2 = self.coord_to_sortable(X2)
        PHI = self.base_model.transition_matrix(t1, t2)
        INTPHI = self.integrated_transition_matrix(t1, t2)[0, :]
        PHIAUG = jnp.eye(self.dimension)
        PHIAUG = PHIAUG.at[: self.d, : self.d].set(PHI)
        for i in range(self.num_insts):
            PHIAUG = PHIAUG.at[self.d + i : self.d + i + 1, : self.d].set(INTPHI)
        return PHIAUG

    def integrated_process_noise(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        """
        Computes the submatrices Qaug12, Qaug21, and Qaug22
        needed to assemble the augmented process noise matrix.

        By default uses the Van Loan method to compute these submatrices.
        Overload this method if you wish to define these submatrices analytically.
        """

        t1 = self.coord_to_sortable(X1)
        t2 = self.coord_to_sortable(X2)
        dt = t2 - t1
        F = self.base_model.design_matrix()
        L = self.base_model.noise_effect_matrix()
        Qc = self.base_model.noise()

        vanloan = smolgp.helpers.VanLoan(F, L, Qc, dt)
        F3 = vanloan["F3"]
        H2 = vanloan["H2"]
        K1 = vanloan["K1"]

        M = F3.T @ H2
        F3TK1 = F3.T @ K1
        W = F3TK1 + F3TK1.T

        Qaug12 = M[:, :1]
        Qaug21 = Qaug12.T
        Qaug22 = W[:1, :1]
        return Qaug12, Qaug21, Qaug22

    # @partial(
    #     jax.jit,
    #     static_argnames=("force_numerical"),
    # )
    def process_noise(
        self, X1: JAXArray, X2: JAXArray, force_numerical: bool = False
    ) -> JAXArray:
        """
        The augmented process noise matrix $Q_k$

        Default behavior computes Q from the Van Loan
        matrix exponential involving F, L, and Qc

        Overload this method if you wish to define the
        integrated process noise analytically.
        """

        t1 = self.coord_to_sortable(X1)
        t2 = self.coord_to_sortable(X2)
        dt = t2 - t1
        if force_numerical:
            Qaug12, Qaug21, Qaug22 = super(type(self), self).integrated_process_noise(
                X1, X2
            )
        else:
            Qaug12, Qaug21, Qaug22 = self.integrated_process_noise(X1, X2)
        Qbase = self.base_model.process_noise(0, dt)
        QAUG = jnp.tile(Qaug22, (self.dimension, self.dimension))
        QAUG = QAUG.at[: self.d, : self.d].set(Qbase)
        for i in range(self.num_insts):
            QAUG = QAUG.at[: self.d, self.d + i : self.d + i + 1].set(Qaug12)
            QAUG = QAUG.at[self.d + i : self.d + i + 1, : self.d].set(Qaug21)
        return QAUG

    def reset_matrix(self, instid: int = 0) -> JAXArray:
        """
        The reset matrix, RESET_k,for instrument `instid` (0-indexed)

        By default, resets only the integral states to zero.
        Overload this method if you wish to define a different reset behavior.
        """
        diag = jnp.ones(self.dimension)
        diag = jax.lax.dynamic_update_slice(diag, jnp.array([0.0]), (self.d + instid,))
        return jnp.diag(diag)


class IntegratedSHO(IntegratedStateSpaceModel):
    r"""The damped, driven simple harmonic oscillator kernel
        integrated over a finite time range, :math:`\delta`.

    This form of the kernel was introduced by `Luhn et al. (in prep)
    <LINK>`_, and it takes the form:

    Args:
        omega: The parameter :math:`\omega_0`.
        quality: The parameter :math:`Q`.
        sigma (optional): The parameter :math:`\sigma`. Defaults to a value of
            1. Specifying the explicit value here provides a slight performance
            boost compared to independently multiplying the kernel with a
            prefactor.
    """

    omega: JAXArray | float
    quality: JAXArray | float
    sigma: JAXArray | float = eqx.field(default_factory=lambda: jnp.ones(()))
    eta: JAXArray | float

    def __init__(
        self,
        omega: JAXArray | float,
        quality: JAXArray | float,
        sigma: JAXArray | float = jnp.ones(()),
        num_insts: int = 1,
        name: str = "IntegratedSHO",
        **kwargs,
    ):
        self.num_insts = num_insts
        self.name = name

        # SHO parameters
        self.omega = omega
        self.quality = quality
        self.sigma = sigma
        self.eta = jnp.sqrt(jnp.abs(1 - 1 / (4 * self.quality**2)))

        # Base model
        self.base_model = smolgp.kernels.SHO(
            omega=self.omega, quality=self.quality, sigma=self.sigma
        )

    def integrated_transition_matrix(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        """The integrated transition matrix Phibar for the SHO process"""

        # Shorthand notations
        n = self.eta
        w = self.omega
        q = self.quality
        a = -0.5 * w / q
        b = n * w
        a2plusb2 = jnp.square(a) + jnp.square(b)
        A = 1 / (2 * n * q)
        B = 1 / (n * w)  # = 1/b
        C = -w / n

        def critical(t1: JAXArray, t2: JAXArray) -> JAXArray:
            ## TODO: returning numerical result until we do this integral by hand
            F = self.base_model.design_matrix()
            return Phibar_from_VanLoan(F, t2 - t1)

        def underdamped(t1: JAXArray, t2: JAXArray) -> JAXArray:
            ## General integral from t1->t2:
            # def Int_ecos(t):
            #     return jnp.exp(a * t) * (a * jnp.cos(b * t) + b * jnp.sin(b * t))
            # def Int_esin(t):
            #     return jnp.exp(a * t) * (a * jnp.sin(b * t) - b * jnp.cos(b * t))
            # Ic = Int_ecos(t2) - Int_ecos(t1)
            # Is = Int_esin(t2) - Int_esin(t1)
            # Phibar11 = Ic + A * Is
            # Phibar12 = B * Is
            # Phibar21 = C * Is
            # Phibar22 = Ic - A * Is

            ## Paper version: hardcoded for t1=0, dt=t2-t1
            dt = t2 - t1
            arg = b * dt
            exp = jnp.exp(a * dt)
            sin = jnp.sin(arg)
            cos = jnp.cos(arg)
            asin = a * sin
            bsin = b * sin
            acos = a * cos
            bcos = b * cos
            Ic = acos + bsin
            Is = asin - bcos
            Phibar11 = exp * (Ic + A * Is) - (a - A * b)
            Phibar12 = B * (Is * exp + b)
            Phibar21 = C * (Is * exp + b)
            Phibar22 = exp * (Ic - A * Is) - (a + A * b)
            return jnp.array([[Phibar11, Phibar12], [Phibar21, Phibar22]]) / a2plusb2

        def overdamped(t1: JAXArray, t2: JAXArray) -> JAXArray:
            ## TODO: returning numerical result until we do this integral by hand
            F = self.base_model.design_matrix()
            return Phibar_from_VanLoan(F, t2 - t1)

        # Return the appropriate form based on quality factor
        t1 = self.coord_to_sortable(X1)
        t2 = self.coord_to_sortable(X2)

        return jax.lax.cond(
            jnp.allclose(q, 0.5),
            critical,
            lambda t1, t2: jax.lax.cond(q > 0.5, underdamped, overdamped, t1, t2),
            t1,
            t2,
        )

    def integrated_process_noise(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        """The integrated process noise submatrices for the SHO process"""
        t1 = self.coord_to_sortable(X1)
        t2 = self.coord_to_sortable(X2)
        dt = t2 - t1
        n = self.eta
        w = self.omega
        q = self.quality
        a = -0.5 * w / q
        b = n * w
        sigma2 = jnp.square(self.sigma)
        A = 1 / (2 * n * q)

        def critical(dt: JAXArray) -> JAXArray:
            # TODO: returning numerical result until we do this integral by hand
            return super(type(self), self).integrated_process_noise(0, dt)

        def underdamped(dt: JAXArray) -> JAXArray:
            x = a * dt
            arg = b * dt
            w2 = jnp.square(w)
            q2 = jnp.square(q)
            q4 = jnp.square(q2)
            exp = jnp.exp(x)
            exp2 = jnp.exp(2 * x)
            exp2m1 = jnp.expm1(2 * x)
            sin = jnp.sin(arg)
            cos = jnp.cos(arg)
            sinsq = jnp.square(jnp.sin(arg))
            sin2 = jnp.sin(2 * arg)
            cos2 = jnp.cos(2 * arg)
            A2 = jnp.square(A)
            iQ12_1 = jnp.square(exp * (cos + A * sin) - 1) / (q * w)
            iQ12_2 = A * exp * (4 * sin - exp * sin2) - 2 * A2 * exp2 * sinsq + exp2m1

            part1 = 8 * q * w * dt + 4 * q2 - 12
            part2 = A2 * exp2 * (cos2 - 16 * q4)
            part3_1 = 16 * exp * (cos + (1 - 2 * q2) * A * sin)
            part3_2 = exp2 * ((1 - 3 * A2) / A * sin2 - 3 * cos2)
            part3 = part3_1 + part3_2
            iQ22 = 1 / (4 * q2 * w2) * (part1 + part2 + part3)
            iQ22 = jnp.maximum(iQ22, 0.0)  # prevent underflows at dt=0

            Qaug12 = sigma2 * jnp.array([[iQ12_1], [iQ12_2]])
            Qaug22 = sigma2 * jnp.array([[iQ22]])
            # Prevent underflows
            Qaug12 = jnp.where(jnp.abs(Qaug12) < 1e-14, jnp.zeros_like(Qaug12), Qaug12)
            Qaug22 = jnp.where(jnp.abs(Qaug22) < 1e-14, jnp.zeros_like(Qaug22), Qaug22)
            Qaug21 = Qaug12.T
            return Qaug12, Qaug21, Qaug22

        def overdamped(dt: JAXArray) -> JAXArray:
            ## TODO: returning numerical result until we do this integral by hand
            return super(type(self), self).integrated_process_noise(0, dt)

        return jax.lax.cond(
            jnp.allclose(q, 0.5),
            critical,
            lambda dt: jax.lax.cond(q > 0.5, underdamped, overdamped, dt),
            dt,
        )


## TODO: is there a way to automate this? aka make a generic IntegratedKernel class...
## Default constructions for all kernels in smolgp.kernels.base
## IntegratedStateSpaceModel parent class will handle the augmentation
## All component matrices will be auto-generated numerically (e.g. A via expm, Q via Van Loan)
class IntegratedExp(IntegratedStateSpaceModel):
    scale: JAXArray | float
    sigma: JAXArray | float = eqx.field(default_factory=lambda: jnp.ones(()))
    lam: JAXArray | float

    def __init__(
        self,
        scale: JAXArray | float,
        sigma: JAXArray | float = jnp.ones(()),
        num_insts: int = 1,
        name: str = "IntegratedExp",
        **kwargs,
    ):
        self.scale = scale
        self.sigma = sigma
        self.name = name
        self.num_insts = num_insts
        self.base_model = smolgp.kernels.Exp(scale=self.scale, sigma=self.sigma)
        self.lam = self.base_model.lam


class IntegratedMatern32(IntegratedStateSpaceModel):
    scale: JAXArray | float
    sigma: JAXArray | float = eqx.field(default_factory=lambda: jnp.ones(()))
    lam: JAXArray | float

    def __init__(
        self,
        scale: JAXArray | float,
        sigma: JAXArray | float = jnp.ones(()),
        num_insts: int = 1,
        name: str = "IntegratedMatern32",
        **kwargs,
    ):
        self.scale = scale
        self.sigma = sigma
        self.name = name
        self.num_insts = num_insts
        self.base_model = smolgp.kernels.Matern32(scale=self.scale, sigma=self.sigma)
        self.lam = self.base_model.lam


class IntegratedMatern52(IntegratedStateSpaceModel):
    scale: JAXArray | float
    sigma: JAXArray | float = eqx.field(default_factory=lambda: jnp.ones(()))
    lam: JAXArray | float

    def __init__(
        self,
        scale: JAXArray | float,
        sigma: JAXArray | float = jnp.ones(()),
        num_insts: int = 1,
        name: str = "IntegratedMatern52",
        **kwargs,
    ):
        self.scale = scale
        self.sigma = sigma
        self.name = name
        self.num_insts = num_insts
        self.base_model = smolgp.kernels.Matern52(scale=self.scale, sigma=self.sigma)
        self.lam = self.base_model.lam


class IntegratedCosine(IntegratedStateSpaceModel):
    scale: JAXArray | float
    sigma: JAXArray | float = eqx.field(default_factory=lambda: jnp.ones(()))
    omega: JAXArray | float

    def __init__(
        self,
        scale: JAXArray | float,
        sigma: JAXArray | float = jnp.ones(()),
        num_insts: int = 1,
        name: str = "IntegratedCosine",
        **kwargs,
    ):
        self.scale = scale
        self.sigma = sigma
        self.name = name
        self.num_insts = num_insts
        self.base_model = smolgp.kernels.Cosine(scale=self.scale, sigma=self.sigma)
        self.omega = self.base_model.omega



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# import smolgp

extensions = [
    "autoapi.extension",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_design",
    "sphinx_togglebutton",
    "sphinxcontrib.lightbox2",
    "myst_nb",
    "IPython.sphinxext.ipython_console_highlighting",
]
myst_enable_extensions = ["amsmath", "dollarmath", "colon_fence"]
math_numfig = True  # for section-numbered equations
master_doc = "index"
source_suffix = {
    ".rst": "restructuredtext",
    ".ipynb": "myst-nb",
}
templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Jupyter -------------------------------------------------
# nb_execution_mode = "auto" # TODO: turn this back on for CI
nb_execution_mode = "off"
nb_execution_excludepatterns = ["benchmarks.ipynb"]
nb_execution_timeout = -1

# -- AutoAPI -------------------------------------------------
autoapi_type = "python"
autoapi_dirs = ["../src"]
# autodoc_output_dir = "apidocs"
# autoapi_add_toctree_entry = True
# autodoc_type_aliases = {
#     "JAXArray": "tinygp.helpers.JAXArray",
# }

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
project = "smolgp"
copyright = "2025, 2026 Ryan Rubenzahl, Soichiro Hattori, Simons Foundation, Inc."
author = "Ryan Rubenzahl & Soichiro Hattori"
# version = smolgp.__version__
# release = smolgp.__version__
version = "0.0.1"
release = "0.0.1"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_book_theme"
html_static_path = ["_static"]
html_title = "smolgp"
html_logo = "_static/smolgp-logo.png"
html_favicon = "_static/favicon.png"
html_show_sourcelink = False
html_theme_options = {
    "path_to_docs": "docs",
    "repository_url": "https://github.com/smolgp-dev/smolgp/",
    "repository_branch": "main",
    # "launch_buttons": {
    #     "binderhub_url": "https://mybinder.org",
    #     "notebook_interface": "jupyterlab",
    #     "colab_url": "https://colab.research.google.com/",
    # },
    "use_edit_page_button": True,
    "use_issues_button": True,
    "use_repository_button": True,
    "use_download_button": True,
}
html_baseurl = "https://smolgp.readthedocs.io/en/latest/"



import jax
import jax.numpy as jnp
import tinygp
import smolgp

from tests.utils import generate_data, generate_integrated_data
from tests.test_kernels import (
    likelihood,
    condition,
    predict,
)

key = jax.random.PRNGKey(0)
jax.config.update("jax_enable_x64", True)


def test_parallel():
    ## SHO Kernel
    S = 2.36
    w = 0.0195
    Q = 7.63
    sigma = jnp.sqrt(S * w * Q)

    ## Base kernels
    kernel_smol = smolgp.kernels.SHO(omega=w, quality=Q, sigma=sigma)
    kernel_tiny = tinygp.kernels.quasisep.SHO(omega=w, quality=Q, sigma=sigma)

    print("Testing ParallelStateSpaceSolver...")
    ## Generate mock data
    N = 50
    yerr = 0.3
    t_train, y_train = generate_data(N, kernel_tiny, yerr=yerr)
    yerr_train = jnp.full_like(t_train, yerr)

    # Build GP objects
    gp_smol = smolgp.GaussianProcess(
        kernel=kernel_smol,
        X=t_train,
        diag=yerr_train**2,
        solver=smolgp.solvers.ParallelStateSpaceSolver,
    )
    gp_tiny = tinygp.GaussianProcess(kernel=kernel_tiny, X=t_train, diag=yerr_train**2)

    # Check likelihood, condition, predict
    likelihood(gp_smol, gp_tiny, y_train, tol=1e-10, atol=1e-13)
    condition(gp_smol, gp_tiny, y_train, tol=1e-9, atol=1e-12)
    predict(gp_smol, gp_tiny, y_train, tol=1e-9, atol=1e-12)

    print()
    print("Testing ParallelIntegratedStateSpaceSolver...")

    ## Mock integrated data
    texp, readout = 140.0, 40.0
    t_train, y_train = generate_integrated_data(
        N, kernel_tiny, texp=texp, readout=readout, yerr=yerr
    )
    texp_train = jnp.full_like(t_train, texp)
    yerr_train = jnp.full_like(t_train, yerr)
    instid = jnp.full_like(t_train, 0).astype(int)  # has to be integer
    X_train = (t_train, texp_train, instid)

    # Integrated kernels
    ikernel_smol = smolgp.kernels.integrated.IntegratedSHO(
        omega=w, quality=Q, sigma=sigma, num_insts=1
    )
    ikernel_tiny = smolgp.kernels.dense.IntegratedSHOKernel(S=S, w=w, Q=Q)

    # Build GP objects
    gp_smol = smolgp.GaussianProcess(
        kernel=ikernel_smol,
        X=X_train,
        diag=yerr_train**2,
        solver=smolgp.solvers.ParallelIntegratedStateSpaceSolver,
    )
    gp_tiny = tinygp.GaussianProcess(kernel=ikernel_tiny, X=X_train, diag=yerr_train**2)

    # Check likelihood, condition, predict
    likelihood(gp_smol, gp_tiny, y_train, tol=1e-10, atol=1e-13)
    condition(gp_smol, gp_tiny, y_train, tol=1e-9, atol=1e-12)
    predict(gp_smol, gp_tiny, y_train, tol=1e-9, atol=1e-12)


if __name__ == "__main__":
    test_parallel()
    print("All parallel solver tests passed.")



"""
These kernels are compatible with :class:`smolgp.solvers.StateSpaceSolver`,
which uses Bayesian filtering and smoothing algorithms to perform scalable GP
inference. (see :ref:`api-solvers-statespace` for more technical details).
On GPU, a performance boost may be observed for large datasets by using the
:class:`smolgp.solvers.parallel.ParallelStateSpaceSolver` class.

Like the quasisep kernels, these methods are experimental, so you may find
the documentation patchy in places. You are encouraged to `open issues or
pull requests <https://github.com/rrubenza/smolgp/issues>`_ as you find gaps.
"""

from __future__ import annotations

import warnings
from abc import abstractmethod
from typing import Any

import equinox as eqx
import jax
import jax.numpy as jnp
from jax.scipy.linalg import expm
from jax.scipy.special import gammaln, factorial

from tinygp.helpers import JAXArray
from tinygp.kernels.base import Kernel
from tinygp.solvers.quasisep.block import Block

from smolgp.helpers import Q_from_VanLoan


def extract_leaf_kernels(kernel, all=False):
    """
    Recursively extract leaf kernels from a sum or product of kernels

    If all==True, extract from both Sum and Product nodes
        (useful for returning all kernel elements).
    Default is False, which only extracts from Sum nodes.
        (as used in decomposing multi-component models, only valid for sums)
    """
    leaf_level = (Sum, Product) if all else (Sum)
    if isinstance(kernel, leaf_level):
        return extract_leaf_kernels(kernel.kernel1, all=all) + extract_leaf_kernels(
            kernel.kernel2, all=all
        )
    else:
        return [kernel]


class StateSpaceModel(Kernel):
    """
    The base class for an instantaneous linear Gaussian state space model

    Has attributes
        `name` (str) : used for unique identification in multicomponent models
        `dimension` (int) : dimensionality $d$ of the state space model

    The components of a state space model are:
    1. design_matrix         : The feedback matrix, F
    2. stationary_covariance : The stationary covariance, Pinf
    3. observation_model     : The observation model, H
    4. noise                 : The spectral density of the white noise process, Qc
    5. noise_effect_matrix   : The noise effect matrix, L
    6. transition_matrix     : The transition matrix, A_k
        (optional, default uses jax.scipy.linalg.expm)
    7. process_noise        : The process noise, Q_k
        (optional, default uses Pinf and A or (alternatively) Van Loan matrix exponential)

    As a child of :class:`tinygp.kernels.Kernel`, this class also implements
    addition and multiplication with other kernels, as well as evaluation

    """

    name: str = eqx.field(static=True)

    def coord_to_sortable(self, X: JAXArray) -> JAXArray:
        """A helper function used to convert coordinates to sortable 1-D values

        By default, this is the identity, but in cases where ``X`` is structured
        (e.g. multivariate inputs), this can be used to appropriately unwrap
        that structure.
        """
        if isinstance(X, tuple):
            return X[0]
        return X

    @property
    def dimension(self) -> int:
        """The dimension of the state space model, d"""
        return self.design_matrix().shape[0]

    @abstractmethod
    def design_matrix(self) -> JAXArray:
        """The design (also called the feedback) matrix for the process, $F$"""
        raise NotImplementedError

    @abstractmethod
    def stationary_covariance(self) -> JAXArray:
        """The stationary covariance of the process, Pinf"""
        raise NotImplementedError

    def observation_model(self, X: JAXArray, component: str | None = None) -> JAXArray:
        """The observation model for the process, $H$"""
        keep = (component is None) or (self.name == component)
        return self.observation_matrix(X) * int(keep)

    @abstractmethod
    def observation_matrix(self, X: JAXArray) -> JAXArray:
        """The observation matrix for the process, $H$"""
        raise NotImplementedError

    @abstractmethod
    def noise(self) -> JAXArray:
        """The spectral density of the white noise process, $Q_c$"""
        raise NotImplementedError

    @abstractmethod
    def noise_effect_matrix(self) -> JAXArray:
        """The noise effect matrix, $L$"""
        raise NotImplementedError

    def transition_matrix(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        """
        The transition matrix between two states at coordinates X1 and X2, $A_k$

        Default behavior uses jax.scipy.linalg.expm(self.design_matrix() * (X2 - X1)),
        which is appropriate for stationary kernels defined by a linear Gaussian SSM.

        Overload this method if you have a more general model or simply wish to
        define the transition matrix analytically.
        """
        F = self.design_matrix()
        t1 = self.coord_to_sortable(X1)
        t2 = self.coord_to_sortable(X2)
        dt = t2 - t1
        return expm(F * dt)

    def process_noise(self, X1: JAXArray, X2: JAXArray, use_van_loan=False) -> JAXArray:
        """
        The process noise matrix $Q_k$

        Default behavior computes Q from Pinf - A @ Pinf @ A
        (see Eq. 7 in Solin & Sarkka 2014). Alternatively,
        give use_van_loan=True to use the Van Loan method to compute
        Q from the matrix exponential involving the F, L, and Qc

        Overload this method if you have a more general model or simply wish to
        define the process noise analytically.
        """
        if use_van_loan:
            # use Van Loan matrix exponential given F, L, Qc
            t1 = self.coord_to_sortable(X1)
            t2 = self.coord_to_sortable(X2)
            dt = t2 - t1
            F = self.design_matrix()
            L = self.noise_effect_matrix()
            Qc = self.noise()
            return Q_from_VanLoan(F, L, Qc, dt)
        else:
            # See Eq. 7 in Solin & Sarkka 2014
            # https://users.aalto.fi/~ssarkka/pub/solin_mlsp_2014.pdf
            Pinf = self.stationary_covariance()
            A = self.transition_matrix(X1, X2)
            return Pinf - A @ Pinf @ A.T

    def reset_matrix(self, instid: int = 0) -> JAXArray:
        """
        The reset matrix for an instantaneous state
        space model is trivially the identity matrix.
        """
        return jnp.eye(self.dimension)

    def __add__(self, other: Kernel | JAXArray) -> Kernel:
        if not isinstance(other, StateSpaceModel):
            raise ValueError(
                "StateSpaceModel kernels can only be added to other StateSpaceModel kernels"
            )
        return Sum(self, other)

    def __radd__(self, other: Any) -> Kernel:
        # We'll hit this first branch when using the `sum` function
        if other == 0:
            return self
        if not isinstance(other, StateSpaceModel):
            raise ValueError(
                "StateSpaceModel kernels can only be added to other StateSpaceModel kernels"
            )
        return Sum(other, self)

    def __mul__(self, other: Kernel | JAXArray) -> Kernel:
        if isinstance(other, StateSpaceModel):
            return Product(self, other)
        if isinstance(other, Kernel) or jnp.ndim(other) != 0:
            raise ValueError(
                "StateSpaceModel kernels can only be multiplied by scalars and other "
                "StateSpaceModel kernels"
            )
        return Scale(kernel=self, scale=other)

    def __rmul__(self, other: Any) -> Kernel:
        if isinstance(other, StateSpaceModel):
            return Product(other, self)
        if isinstance(other, Kernel) or jnp.ndim(other) != 0:
            raise ValueError(
                "StateSpaceModel kernels can only be multiplied by scalars and other "
                "StateSpaceModel kernels"
            )
        return Scale(kernel=self, scale=other)

    def evaluate(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        """
        The kernel evaluated via the state space representatio

        See Eq. 4 in Hartikainen & Särkkä 2010
        https://users.aalto.fi/~ssarkka/pub/gp-ts-kfrts.pdf
        """
        Pinf = self.stationary_covariance()
        h1 = self.observation_model(X1)
        h2 = self.observation_model(X2)
        return jnp.where(
            self.coord_to_sortable(X1) < self.coord_to_sortable(X2),
            (h2 @ Pinf @ self.transition_matrix(X1, X2).T @ h1.T).squeeze(),
            (h1 @ self.transition_matrix(X2, X1) @ Pinf @ h2.T).squeeze(),
        )

    def evaluate_diag(self, X: JAXArray) -> JAXArray:
        """For state space kernels, the variance is simple to compute"""
        h = self.observation_model(X)
        return h @ self.stationary_covariance() @ h.T

    def psd(self, omega: JAXArray) -> JAXArray:
        """
        The power spectral density (PSD) of the kernel

        See Eq. 8 in Solin & Sarkka 2014
        https://users.aalto.fi/~ssarkka/pub/solin_mlsp_2014.pdf
        """
        F = self.design_matrix()
        L = self.noise_effect_matrix()
        Qc = self.noise()
        H = self.observation_matrix(0)  # PSD is stationary, so X doesn't matter
        I = jnp.eye(self.dimension)

        def compute_psd(w: JAXArray) -> JAXArray:
            M = jnp.linalg.inv(F + 1j * w * I)
            S = H @ M.conj() @ L @ Qc @ L.T @ M.T @ H.T
            return S.squeeze().real / (2 * jnp.pi)

        return jax.vmap(compute_psd)(omega)


class Sum(StateSpaceModel):
    """
    A helper to represent the sum of two StateSpaceModel kernels

    The state dimension becomes d = d1 + d2
    """

    kernel1: StateSpaceModel
    kernel2: StateSpaceModel

    def __init__(self, kernel1, kernel2):
        self.kernel1 = kernel1
        self.kernel2 = kernel2
        self.name = f"Sum({kernel1.name}, {kernel2.name})"

    @property
    def dimension(self) -> int:
        """The dimension of the summed state space model"""
        return self.kernel1.dimension + self.kernel2.dimension

    def coord_to_sortable(self, X: JAXArray) -> JAXArray:
        """We assume that both kernels use the same coordinates"""
        return self.kernel1.coord_to_sortable(X)

    def design_matrix(self) -> JAXArray:
        """F = BlockDiag(F1, F2)"""
        return Block(
            self.kernel1.design_matrix(),
            self.kernel2.design_matrix(),
        ).to_dense()

    def noise_effect_matrix(self) -> JAXArray:
        """L = BlockDiag(L1, L2)"""
        return Block(
            self.kernel1.noise_effect_matrix(), self.kernel2.noise_effect_matrix()
        ).to_dense()

    def stationary_covariance(self) -> JAXArray:
        """Pinf = BlockDiag(Pinf1, Pinf2)"""
        return Block(
            self.kernel1.stationary_covariance(),
            self.kernel2.stationary_covariance(),
        ).to_dense()

    def transition_matrix(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        """A = BlockDiag(A1, A2)"""
        return Block(
            self.kernel1.transition_matrix(X1, X2),
            self.kernel2.transition_matrix(X1, X2),
        ).to_dense()

    def process_noise(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        """Q = BlockDiag(Q1, Q2)"""
        return Block(
            self.kernel1.process_noise(X1, X2), self.kernel2.process_noise(X1, X2)
        ).to_dense()

    def observation_model(self, X: JAXArray, component=None) -> JAXArray:
        """H = [H1, H2] with component extraction"""
        return jnp.hstack(
            (
                self.kernel1.observation_model(X, component=component),
                self.kernel2.observation_model(X, component=component),
            )
        )

    def observation_matrix(self, X: JAXArray) -> JAXArray:
        """H = [H1, H2]"""
        return jnp.hstack(
            (
                self.kernel1.observation_matrix(X),
                self.kernel2.observation_matrix(X),
            )
        )

    def reset_matrix(self, instid: int = 0) -> JAXArray:
        """RESET = BlockDiag(RESET1, RESET2)"""
        return Block(
            self.kernel1.reset_matrix(instid), self.kernel2.reset_matrix(instid)
        ).to_dense()

    def noise(self) -> JAXArray:
        """Qc = BlockDiag(Qc1, Qc2)"""
        return Block(self.kernel1.noise(), self.kernel2.noise()).to_dense()


class Product(StateSpaceModel):
    """
    A helper to represent the product of two StateSpaceModel kernels

    The state dimension becomes d = d1 * d2
    """

    kernel1: StateSpaceModel
    kernel2: StateSpaceModel

    def __init__(self, kernel1, kernel2):
        self.kernel1 = kernel1
        self.kernel2 = kernel2
        self.name = f"Product({kernel1.name}, {kernel2.name})"

    @property
    def dimension(self) -> int:
        """The dimension of the product state space model"""
        return self.kernel1.dimension * self.kernel2.dimension

    def coord_to_sortable(self, X: JAXArray) -> JAXArray:
        """We assume that both kernels use the same coordinates"""
        return self.kernel1.coord_to_sortable(X)

    def design_matrix(self) -> JAXArray:
        """F = F1 ⊗ I + I ⊗ F2"""
        F1 = self.kernel1.design_matrix()
        F2 = self.kernel2.design_matrix()
        I1 = jnp.eye(F1.shape[0])
        I2 = jnp.eye(F2.shape[0])
        return jnp.kron(F1, I2) + jnp.kron(I1, F2)

    def noise_effect_matrix(self) -> JAXArray:
        """
        L for products is not uniquely defined!
        We choose a convenient form here where
        L is simply the identity, and choose
        Qc such that we get the correct L@Qc@L^T
        """
        return jnp.eye(self.dimension)

    def stationary_covariance(self) -> JAXArray:
        """Pinf = Pinf1 ⊗ Pinf2"""
        Pinf1 = self.kernel1.stationary_covariance()
        Pinf2 = self.kernel2.stationary_covariance()
        return jnp.kron(Pinf1, Pinf2)

    def transition_matrix(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        """A = A1 ⊗ A2"""
        A1 = self.kernel1.transition_matrix(X1, X2)
        A2 = self.kernel2.transition_matrix(X1, X2)
        return jnp.kron(A1, A2)

    def process_noise(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        """
        Q for a product is best determined via the identity
        Q = Pinf - A Pinf A^T
        """
        Q1 = self.kernel1.process_noise(X1, X2)
        Q2 = self.kernel2.process_noise(X1, X2)
        Pinf1 = self.kernel1.stationary_covariance()
        Pinf2 = self.kernel2.stationary_covariance()
        return jnp.kron(Pinf1, Q2) + jnp.kron(Q1, Pinf2) - jnp.kron(Q1, Q2)

    def observation_model(self, X: JAXArray, component=None) -> JAXArray:
        """H = H1 ⊗ H2 with component extraction"""
        H1 = self.kernel1.observation_model(X, component=component)
        H2 = self.kernel2.observation_model(X, component=component)
        return jnp.kron(H1, H2)

    def observation_matrix(self, X: JAXArray) -> JAXArray:
        """H = H1 ⊗ H2"""
        H1 = self.kernel1.observation_matrix(X)
        H2 = self.kernel2.observation_matrix(X)
        return jnp.kron(H1, H2)

    def reset_matrix(self, instid: int = 0) -> JAXArray:
        """RESET = RESET1 ⊗ RESET2"""
        Reset1 = self.kernel1.reset_matrix(instid)
        Reset2 = self.kernel2.reset_matrix(instid)
        return jnp.kron(Reset1, Reset2)

    def noise(self) -> JAXArray:
        """
        Qc for products is not uniquely defined!
        Here we choose a convenient form where
        Qc = L1 Qc1 L1^T ⊗ Pinf2 + Pinf1 ⊗ L2 Qc2 L2^T
        and L is the identity, so that L Qc L^T gives the correct process noise
        """
        Qc1 = self.kernel1.noise()
        Qc2 = self.kernel2.noise()
        L1 = self.kernel1.noise_effect_matrix()
        L2 = self.kernel2.noise_effect_matrix()
        Pinf1 = self.kernel1.stationary_covariance()
        Pinf2 = self.kernel2.stationary_covariance()
        B1 = L1 @ Qc1 @ L1.T
        B2 = L2 @ Qc2 @ L2.T
        B = jnp.kron(B1, Pinf2) + jnp.kron(Pinf1, B2)
        return B


class Wrapper(StateSpaceModel):
    """A base class for wrapping kernels with some custom implementations"""

    kernel: StateSpaceModel

    @property
    def dimension(self) -> int:
        return self.kernel.dimension

    def coord_to_sortable(self, X: JAXArray) -> JAXArray:
        return self.kernel.coord_to_sortable(X)

    def design_matrix(self) -> JAXArray:
        return self.kernel.design_matrix()

    def noise_effect_matrix(self) -> JAXArray:
        return self.kernel.noise_effect_matrix()

    def noise(self) -> JAXArray:
        return self.kernel.noise()

    def stationary_covariance(self) -> JAXArray:
        return self.kernel.stationary_covariance()

    def observation_matrix(self, X: JAXArray) -> JAXArray:
        return self.kernel.observation_matrix(X)

    def transition_matrix(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        return self.kernel.transition_matrix(X1, X2)

    def process_noise(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        return self.kernel.process_noise(X1, X2)

    def reset_matrix(self, instid=0):
        return self.kernel.reset_matrix(instid)


class Scale(Wrapper):
    """The product of a scalar and a quasiseparable kernel"""

    scale: JAXArray | float

    def stationary_covariance(self) -> JAXArray:
        return self.scale * self.kernel.stationary_covariance()

    # TODO: also scale Qc?
    def noise(self) -> JAXArray:
        return self.scale * self.kernel.noise()


################ GP KERNEL DEFINITIONS ################
## TODO: tinygp base kernels not yet implemented
##       some will require approximations
# Polynomial
# ExpSquared (RBF), will need approx. via Taylor expannsion
#     see Hartikainen and Särkkä, 2010; Särkkä et al., 2013
# ExpSineSquared (will need approx)
# RationalQuadratic
# Quasiperiodic (not explicitly in tinygp, but is: ExpSquared * ExpSineSquared
##    some also define ExpSquared * ExpSineSquared * ExpCosineSquared for P/2 term
## RotationTerm (sum of SHO as in celerite)


class Constant(StateSpaceModel):
    r"""A constant kernel

    .. math::

        k(\mathbf{x}_i,\,\mathbf{x}_j) = \sigma^2

    where :math:`\sigma` is a parameter.

    Args:
        sigma: The parameter :math:`\sigma` in the above equation.
    """

    sigma: JAXArray | float

    def __init__(
        self,
        sigma: JAXArray | float = jnp.ones(()),
        name: str = "Constant",
        **kwargs,
    ):
        self.sigma = sigma
        self.name = name

    def design_matrix(self) -> JAXArray:
        """The design (also called the feedback) matrix for a Constant process, F"""
        return jnp.array([[0]])

    def stationary_covariance(self) -> JAXArray:
        """The stationary covariance of a Constant process, Pinf"""
        return jnp.array([[jnp.square(self.sigma)]])

    def noise(self) -> JAXArray:
        """The scalar Qc for a Constant process"""
        return jnp.array([[0]])

    def observation_matrix(self, X: JAXArray) -> JAXArray:
        """The observation model H for a Constant process"""
        del X
        return jnp.array([[1]])

    def noise_effect_matrix(self) -> JAXArray:
        """The noise effect matrix L for a Constant process"""
        return jnp.array([[1]])

    def transition_matrix(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        """The transition matrix A_k for a Constant process"""
        del X1, X2
        return self.eye(self.dimension)

    def process_noise(self, X1, X2, use_van_loan=False):
        """The process noise Q_k for a Constant process"""
        return jnp.array([[0]])


class SHO(StateSpaceModel):
    r"""The damped, driven simple harmonic oscillator kernel

    This form of the kernel was introduced by `Foreman-Mackey et al. (2017)
    <https://arxiv.org/abs/1703.09710>`_, and it takes the form:

    .. math::

        k(\Delta) = \sigma^2\,\exp\left(-\frac{\omega_0\,\Delta}{2\,Q}\right)
        \left\{\begin{array}{ll}
            1 + \omega_0\,\Delta & \mbox{for } Q = 1/2 \\
            \cosh(\eta\,\omega_0\,\Delta) + \fra{1}{2\eta Q} \sinh(\eta\,\omega_0\,\Delta)
                & \mbox{for } Q < 1/2 \\
            \frac{1}{2\eta Q}\cos(\eta\,\omega_0\,\Delta) + \sin(\eta\,\omega_0\,\Delta)
                & \mbox{for } Q > 1/2
        \end{array}\right.

    for :math:`\Delta = |x_i - x_j|`, :math:`\eta = \sqrt{|1 - 1/(4Q^2)|}`.

    Args:
        omega: The parameter :math:`\omega_0`.
        quality: The parameter :math:`Q`.
        sigma (optional): The parameter :math:`\sigma`. Defaults to a value of
            1. Specifying the explicit value here provides a slight performance
            boost compared to independently multiplying the kernel with a
            prefactor.
    """

    omega: JAXArray | float
    quality: JAXArray | float
    sigma: JAXArray | float = eqx.field(default_factory=lambda: jnp.ones(()))
    eta: JAXArray | float

    def __init__(
        self,
        omega: JAXArray | float,
        quality: JAXArray | float,
        sigma: JAXArray | float = jnp.ones(()),
        name: str = "SHO",
        **kwargs,
    ):
        # SHO parameters
        self.omega = omega
        self.quality = quality
        self.sigma = sigma
        self.name = name
        self.eta = jnp.sqrt(jnp.abs(1 - 1 / (4 * self.quality**2)))

    def design_matrix(self) -> JAXArray:
        """The design (also called the feedback) matrix for the SHO process, F"""
        return jnp.array(
            [[0, 1], [-jnp.square(self.omega), -self.omega / self.quality]]
        )

    def stationary_covariance(self) -> JAXArray:
        """The stationary covariance of the SHO process, Pinf"""
        return jnp.square(self.sigma) * jnp.diag(jnp.array([1, jnp.square(self.omega)]))

    def noise(self) -> JAXArray:
        """The scalar Qc for the SHO process"""
        omega3 = jnp.power(self.omega, 3)
        return jnp.array([[2 * omega3 * jnp.square(self.sigma) / self.quality]])

    def observation_matrix(self, X: JAXArray) -> JAXArray:
        """The observation model H for the SHO process"""
        del X
        return jnp.array([[1, 0]])

    def noise_effect_matrix(self) -> JAXArray:
        """The noise effect matrix L for the SHO process"""
        return jnp.array([[0], [1]])

    def transition_matrix(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        """The transition matrix A_k for the SHO process"""
        t1 = self.coord_to_sortable(X1)
        t2 = self.coord_to_sortable(X2)
        dt = t2 - t1
        n = self.eta
        w = self.omega
        q = self.quality

        def critical(dt: JAXArray) -> JAXArray:
            return jnp.exp(-w * dt) * jnp.array(
                [[1 + w * dt, dt], [-jnp.square(w) * dt, 1 - w * dt]]
            )

        def underdamped(dt: JAXArray) -> JAXArray:
            f = 2 * n * q
            x = n * w * dt
            sin = jnp.sin(x)
            cos = jnp.cos(x)
            return jnp.exp(-0.5 * w * dt / q) * jnp.array(
                [[cos + sin / f, sin / (w * n)], [-w * sin / n, cos - sin / f]]
            )

        def overdamped(dt: JAXArray) -> JAXArray:
            f = 2 * n * q
            x = n * w * dt
            sinh = jnp.sinh(x)
            cosh = jnp.cosh(x)
            return jnp.exp(-0.5 * w * dt / q) * jnp.array(
                [[cosh + sinh / f, sinh / (w * n)], [-w * sinh / n, cosh - sinh / f]]
            )

        return jax.lax.cond(
            jnp.allclose(q, 0.5),
            critical,
            lambda dt: jax.lax.cond(q > 0.5, underdamped, overdamped, dt),
            dt,
        )

    def process_noise(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        """The process noise Q_k for the SHO process"""
        t1 = self.coord_to_sortable(X1)
        t2 = self.coord_to_sortable(X2)
        dt = t2 - t1
        n = self.eta
        w = self.omega
        q = self.quality

        def critical(dt: JAXArray) -> JAXArray:
            arg = 2 * w * dt
            argexp = arg * jnp.exp(-arg)
            expm1 = jnp.expm1(-arg)
            wdt = w * dt
            Q11 = -expm1 - argexp * (1 + wdt)
            Q12 = Q21 = w * wdt * argexp
            Q22 = jnp.square(w) * (-expm1 - argexp * (-1 + wdt))
            return jnp.square(self.sigma) * jnp.array([[Q11, Q12], [Q21, Q22]])

        def underdamped(dt: JAXArray) -> JAXArray:
            f = 2 * n * q
            q2 = jnp.square(q)
            n2 = jnp.square(n)
            w2 = jnp.square(w)
            a = w * dt / q  # argument in exponential
            x = n * w * dt  # argument in sin/cos
            sin = jnp.sin(x)
            sin2 = jnp.sin(2 * x)
            sinsq = jnp.square(sin)
            exp = jnp.exp(-a)
            expm1 = jnp.expm1(-a)  # exp(-a) - 1
            Q11 = -expm1 - (sin2 / f + sinsq / (2 * n2 * q2)) * exp
            Q12 = Q21 = exp * (w * sinsq / (n2 * q))
            Q22 = w2 * (-expm1 + exp * (sin2 / f - sinsq / (2 * n2 * q2)))
            Q = jnp.square(self.sigma) * jnp.array([[Q11, Q12], [Q21, Q22]])
            return jnp.where(
                jnp.abs(Q) < 1e-14, jnp.zeros_like(Q), Q
            )  # prevent underflows

        def overdamped(dt: JAXArray) -> JAXArray:
            f = 2 * n * q
            q2 = jnp.square(q)
            n2 = jnp.square(n)
            w2 = jnp.square(w)
            a = w * dt / q  # argument in exponential
            x = n * w * dt  # argument in sin/cos
            sinh = jnp.sinh(x)
            sinh2 = jnp.sinh(2 * x)
            sinhsq = jnp.square(sinh)
            exp = jnp.exp(-a)
            expm1 = jnp.expm1(-a)  # exp(-a) - 1
            Q11 = -expm1 - (sinh2 / f + sinhsq / (2 * n2 * q2)) * exp
            Q12 = Q21 = exp * (w * sinhsq / (n2 * q))
            Q22 = w2 * (-expm1 + exp * (sinh2 / f - sinhsq / (2 * n2 * q2)))
            return jnp.square(self.sigma) * jnp.array([[Q11, Q12], [Q21, Q22]])

        return jax.lax.cond(
            jnp.allclose(q, 0.5),
            critical,
            lambda dt: jax.lax.cond(q > 0.5, underdamped, overdamped, dt),
            dt,
        )


class Exp(StateSpaceModel):
    r"""A state space implementation of :class:`tinygp.kernels.quasisep.Exp`

    This kernel takes the form:

    .. math::

        k(\Delta)=\sigma^2\,\exp\left(-\frac{\Delta}{\ell}\right)

    for :math:`\Delta = |x_i - x_j|`. Also known as the "Ornstein-Uhlenbeck" kernel,
    and is also equivalent to a Matérn-1/2 kernel.

    Args:
        scale: The parameter :math:`\ell`.
        sigma (optional): The parameter :math:`\sigma`. Defaults to a value of 1.
    """

    scale: JAXArray | float
    sigma: JAXArray | float = eqx.field(default_factory=lambda: jnp.ones(()))
    lam: JAXArray | float

    def __init__(
        self,
        scale: JAXArray | float,
        sigma: JAXArray | float = jnp.ones(()),
        name: str = "Exp",
        **kwargs,
    ):
        # Exp parameters
        self.scale = scale
        self.sigma = sigma
        self.name = name
        self.lam = 1 / self.scale

    def design_matrix(self) -> JAXArray:
        """The design (also called the feedback) matrix for the Exp process, F"""
        return jnp.array([[-self.lam]])

    def stationary_covariance(self) -> JAXArray:
        """The stationary covariance of the Exp process, Pinf"""
        return jnp.square(self.sigma) * jnp.ones((1, 1))

    def observation_matrix(self, X: JAXArray) -> JAXArray:
        """The observation model H for the Exp process"""
        del X
        return jnp.array([[1]])

    def noise_effect_matrix(self) -> JAXArray:
        """The noise effect matrix L for the Exp process"""
        return jnp.array([[1]])

    def noise(self) -> JAXArray:
        """The scalar Qc for the Exp process"""
        return jnp.array([[2 * self.lam * jnp.square(self.sigma)]])

    def transition_matrix(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        """The transition matrix A_k for the Exp process"""
        t1 = self.coord_to_sortable(X1)
        t2 = self.coord_to_sortable(X2)
        dt = t2 - t1
        return jnp.exp(-dt[None, None] / self.scale)

    def process_noise(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        """The process noise Q_k for the Exp process"""
        t1 = self.coord_to_sortable(X1)
        t2 = self.coord_to_sortable(X2)
        dt = t2 - t1
        sigma2 = jnp.square(self.sigma)
        return jnp.array([[sigma2 * (1 - jnp.exp(-2 * dt * self.lam))]])


class Matern32(StateSpaceModel):
    r"""A state space implementation of :class:`tinygp.kernels.quasisep.Matern32`

    This kernel takes the form:

    .. math::

        k(\Delta)=\sigma^2\,\left(1+f\,\Delta\right)\,\exp(-f\,\Delta)

    for :math:`\Delta = |x_i - x_j|` and :math:`f = \sqrt{3} / \ell`.

    Args:
        scale: The parameter :math:`\ell`.
        sigma (optional): The parameter :math:`\sigma`. Defaults to a value of 1.
    """

    scale: JAXArray | float
    sigma: JAXArray | float
    lam: JAXArray | float

    def __init__(
        self,
        scale: JAXArray | float,
        sigma: JAXArray | float = jnp.ones(()),
        name: str = "Matern32",
        **kwargs,
    ):
        # Matern-3/2 parameters
        self.scale = scale
        self.sigma = sigma
        self.name = name
        self.lam = jnp.sqrt(3) / self.scale

    def design_matrix(self) -> JAXArray:
        """The design (also called the feedback) matrix for the Matern-3/2 process, F"""
        lam2 = jnp.square(self.lam)
        return jnp.array([[0, 1], [-lam2, -2 * self.lam]])

    def stationary_covariance(self) -> JAXArray:
        """The stationary covariance of the Matern-3/2 process, Pinf"""
        return jnp.square(self.sigma) * jnp.diag(jnp.array([1, jnp.square(self.lam)]))

    def observation_matrix(self, X: JAXArray) -> JAXArray:
        """The observation model H for the Matern-3/2 process"""
        del X
        return jnp.array([[1, 0]])

    def noise_effect_matrix(self) -> JAXArray:
        """The noise effect matrix L for the Matern-3/2 process"""
        return jnp.array([[0], [1]])

    def noise(self) -> JAXArray:
        """The scalar Qc for the Matern-3/2 process"""
        lam3 = jnp.power(self.lam, 3)
        return jnp.array([[4 * lam3 * jnp.square(self.sigma)]])

    def transition_matrix(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        """The transition matrix A_k for the Matern-3/2 process"""
        t1 = self.coord_to_sortable(X1)
        t2 = self.coord_to_sortable(X2)
        dt = t2 - t1
        lam = self.lam
        return jnp.exp(-lam * dt) * jnp.array(
            [[1 + lam * dt, dt], [-jnp.square(lam) * dt, 1 - lam * dt]]
        )

    def process_noise(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        """The process noise Q_k for the Matern-3/2 process"""
        t1 = self.coord_to_sortable(X1)
        t2 = self.coord_to_sortable(X2)
        dt = t2 - t1
        lam = self.lam
        arg = 2 * lam * dt
        d2l3 = 2 * jnp.square(dt) * jnp.power(lam, 3)
        exp = jnp.exp(-arg)
        expm1 = jnp.expm1(-arg)
        return jnp.square(self.sigma) * jnp.array(
            [
                [-expm1 - arg * (lam * dt + 1) * exp, d2l3 * exp],
                [d2l3 * exp, jnp.square(lam) * (-expm1 - arg * (lam * dt - 1) * exp)],
            ]
        )


class Matern52(StateSpaceModel):
    r"""A state space implementation of :class:`tinygp.kernels.quasisep.Matern52`

    This kernel takes the form:

    .. math::

        k(\Delta)=\sigma^2\,\left(1+f\,\Delta + \frac{f^2\,\Delta^2}{3}\right)
            \,\exp(-f\,\Delta)

    for :math:`\Delta = |x_i - x_j|` and :math:`f = \sqrt{5} / \ell`.

    Args:
        scale: The parameter :math:`\ell`.
        sigma (optional): The parameter :math:`\sigma`. Defaults to a value of 1.
    """

    scale: JAXArray | float
    sigma: JAXArray | float
    lam: JAXArray | float

    def __init__(
        self,
        scale: JAXArray | float,
        sigma: JAXArray | float = jnp.ones(()),
        name: str = "Matern52",
        **kwargs,
    ):
        # Matern-5/2 parameters
        self.scale = scale
        self.sigma = sigma
        self.name = name
        self.lam = jnp.sqrt(5) / self.scale

    def design_matrix(self) -> JAXArray:
        """The design (also called the feedback) matrix for the Matern-5/2 process, F"""
        lam2 = jnp.square(self.lam)
        lam3 = lam2 * self.lam
        return jnp.array([[0, 1, 0], [0, 0, 1], [-lam3, -3 * lam2, -3 * self.lam]])

    def stationary_covariance(self) -> JAXArray:
        """The stationary covariance of the Matern-5/2 process, Pinf"""
        lam2 = jnp.square(self.lam)
        lam2o3 = lam2 / 3
        return jnp.square(self.sigma) * jnp.array(
            [[1, 0, -lam2o3], [0, lam2o3, 0], [-lam2o3, 0, jnp.square(lam2)]]
        )

    def observation_matrix(self, X: JAXArray) -> JAXArray:
        """The observation model H for the Matern-5/2 process"""
        del X
        return jnp.array([[1, 0, 0]])

    def noise_effect_matrix(self) -> JAXArray:
        """The noise effect matrix L for the Matern-5/2 process"""
        return jnp.array([[0], [0], [1]])

    def noise(self) -> JAXArray:
        """The scalar Qc for the Matern-5/2 process"""
        lam5 = jnp.power(self.lam, 5)
        return jnp.array([[16 * lam5 * jnp.square(self.sigma) / 3]])

    def transition_matrix(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        """The transition matrix A_k for the Matern-5/2 process"""
        t1 = self.coord_to_sortable(X1)
        t2 = self.coord_to_sortable(X2)
        dt = t2 - t1
        lam = self.lam
        lam2 = jnp.square(lam)
        d2 = jnp.square(dt)
        a11 = 0.5 * lam2 * d2 + lam * dt + 1
        a12 = dt * (lam * dt + 1)
        a13 = 0.5 * d2
        a21 = -0.5 * lam * lam2 * d2
        a22 = -lam2 * d2 + lam * dt + 1
        a23 = 0.5 * dt * (2 - lam * dt)
        a31 = 0.5 * lam2 * lam * dt * (lam * dt - 2)
        a32 = lam2 * dt * (lam * dt - 3)
        a33 = 0.5 * lam2 * d2 - 2 * lam * dt + 1
        return jnp.exp(-lam * dt) * jnp.array(
            [
                [a11, a12, a13],
                [a21, a22, a23],
                [a31, a32, a33],
            ]
        )

    def process_noise(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        """The process noise Q_k for the Matern-5/2 process"""
        t1 = self.coord_to_sortable(X1)
        t2 = self.coord_to_sortable(X2)
        dt = t2 - t1
        lam = self.lam
        dtl = dt * lam
        l2 = jnp.square(lam)
        dt2 = jnp.square(dt)
        l4 = jnp.power(lam, 4)
        dt4 = jnp.power(dt, 4)
        l5 = jnp.power(lam, 5)
        dt2l5 = dt2 * l5
        arg = -2 * dtl
        exp = jnp.exp(arg)
        expm1 = jnp.expm1(arg)
        Q11 = -3 * expm1 - 2 * dtl * exp * (3 + dtl * (3 + dtl * (2 + dtl)))
        Q12 = 2 * exp * dt4 * l5
        Q13 = -l2 * (-expm1 + 2 * dtl * exp * (-1 + dtl * (-1 + dtl * (-2 + dtl))))
        Q21 = Q12
        Q22 = l2 * (-expm1 - 2 * dtl * exp * (1 + dtl * jnp.square(1 - dtl)))
        Q23 = 2 * exp * dt2l5 * jnp.square(-2 + dtl)
        Q31 = Q13
        Q32 = Q23
        Q33 = l4 * (-3 * expm1 - 2 * dtl * exp * (-5 + dtl * (11 + dtl * (-6 + dtl))))

        return (jnp.square(self.sigma) / 3) * jnp.array(
            [
                [Q11, Q12, Q13],
                [Q21, Q22, Q23],
                [Q31, Q32, Q33],
            ]
        )


class Cosine(StateSpaceModel):
    r"""A state space implementation of :class:`tinygp.kernels.quasisep.Cosine`

    This kernel takes the form:

    .. math::

        k(\Delta)=\sigma^2\,\cos(-2\,\pi\,\Delta/\ell)

    for :math:`\Delta = |x_i - x_j|`.

    Args:
        scale: The parameter :math:`\ell`.
        sigma (optional): The parameter :math:`\sigma`. Defaults to a value of 1.
    """

    scale: JAXArray | float
    sigma: JAXArray | float = eqx.field(default_factory=lambda: jnp.ones(()))
    omega: JAXArray | float

    def __init__(self, scale, sigma=jnp.ones(()), name="Cosine", **kwargs):
        # Cosine parameters
        self.scale = scale
        self.sigma = sigma
        self.name = name
        self.omega = 2 * jnp.pi / self.scale

    def design_matrix(self) -> JAXArray:
        """The design (also called the feedback) matrix for the Cosine process, F"""
        return jnp.array([[0, -self.omega], [self.omega, 0]])

    def stationary_covariance(self) -> JAXArray:
        """The stationary covariance of the Cosine process, Pinf"""
        return jnp.square(self.sigma) * jnp.eye(2)

    def observation_matrix(self, X: JAXArray) -> JAXArray:
        """The observation model H for the Cosine process"""
        del X
        return jnp.array([[1, 0]])

    def noise_effect_matrix(self) -> JAXArray:
        """The noise effect matrix L for the Cosine process"""
        return jnp.array([[0], [1]])

    def noise(self) -> JAXArray:
        """The scalar Qc for the Cosine process"""
        Qc = 0.0  # Does not have a white noise driving process
        return jnp.array([[Qc]])

    def transition_matrix(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        """The transition matrix A_k for the Cosine process"""
        t1 = self.coord_to_sortable(X1)
        t2 = self.coord_to_sortable(X2)
        dt = t2 - t1
        arg = self.omega * dt
        cos = jnp.cos(arg)
        sin = jnp.sin(arg)
        return jnp.array([[cos, -sin], [sin, cos]])

    def process_noise(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
        """The process noise Q_k for the Cosine process"""
        del X1, X2
        return jnp.zeros((self.dimension, self.dimension))


# class ExpSquared(StateSpaceModel):
#     r"""
#     A state space implementation of the exponential squared
#     (also called the radial basis function, or RBF) kernel

#     .. math::

#         k(\mathbf{x}_i,\,\mathbf{x}_j) = \exp(-r^2 / 2)

#     where, by default,

#     .. math::

#         r^2 = ||(\mathbf{x}_i - \mathbf{x}_j) / \ell||_2^2

#     Args:
#         scale: The parameter :math:`\ell`.
#     """


class ExpSineSquared(Wrapper):
    r"""The exponential sine squared or "periodic" kernel

    .. math::

        k(\mathbf{x}_i,\,\mathbf{x}_j) = \sigma^2 \exp(-\Gamma\,\sin^2 \pi r)

    where, by default,

    .. math::

        r = ||(\mathbf{x}_i - \mathbf{x}_j) / P||_1

    In the state space representation, this kernel is approximated using
    a finite number of basis functions. The method was introduced by
    `Solin & Särkkä (2014) <https://proceedings.mlr.press/v33/solin14.html`_.
    See their Figure 2 for number of basis functions to reach desired accuracy.
    Default behavior will automatically select the number of basis functions
    based on the length scale :math:`\ell`.

    Args:
        period: The parameter :math:`P`.
        gamma: The parameter :math:`\Gamma`.
        sigma: The parameter :math:`\sigma`. Defaults to a value of 1.
    """

    gamma: JAXArray | float | None = None
    period: JAXArray | float
    scale: JAXArray | float
    omega: JAXArray | float
    sigma: JAXArray | float = eqx.field(default_factory=lambda: jnp.ones(()))
    order: int | None
    kernel: StateSpaceModel

    def __init__(
        self,
        period: JAXArray | float,
        gamma: JAXArray | float = jnp.ones(()),
        sigma: JAXArray | float = jnp.ones(()),
        name: str = "ExpSineSquared",
        order: int | None = None,
        **kwargs,
    ):
        self.period = period  # P
        self.gamma = gamma  # \Gamma
        self.sigma = sigma  # \sigma
        self.name = name

        self.scale = 2 / self.gamma  # \ell^2 in Solin & Särkkä (2014)
        self.omega = 2 * jnp.pi / self.period  # \omega_0 in Solin & Särkkä (2014)

        # Auto-select order (J) using Fig 2 of
        # Solin & Särkkä (2014) as a
        ell = jnp.sqrt(self.scale)
        if order is None:
            if ell >= 1:
                self.order = 4
            elif ell >= 0.5:
                self.order = 6
            elif ell >= 0.25:
                self.order = 8
            else:
                self.order = 16
                warnings.warn(
                    "ExpSineSquared kernel with scale < 0.25 (gamma > 16) may require a high order approximation; "
                    "it may be worthwhile to change units to a more compatible scale (recommended) "
                    "or specify the 'order' parameter explicitly."
                )
        else:
            self.order = order

        # Construct the kernel as a sum of PeriodicTerms
        kernel = self.PeriodicTerm(0, self.omega, self.scale)
        for j in range(1, self.order):
            kernel += self.PeriodicTerm(j, self.omega, self.scale)
        self.kernel = kernel

    def error_bound(self):
        """
        An upper bound on the error in the covariance
        from the Taylor series approximation.

        See Sec 3.4 of Solin & Särkkä (2014).
        """
        J = self.order
        return jnp.exp(1 - self.scale) / jax.scipy.special.factorial(J + 1)

    def stationary_covariance(self) -> JAXArray:
        return jnp.square(self.sigma) * self.kernel.stationary_covariance()

    # Class for each periodic term in the Taylor series expansion
    class PeriodicTerm(StateSpaceModel):
        """
        A single term in the state space representation of the
        exponential sine squared or "periodic" kernel

        See Solin & Särkkä (2014) for details.
        """

        order: int
        omega: JAXArray | float
        scale: JAXArray | float

        def __init__(self, order, omega, scale, **kwargs):
            self.order = order
            self.omega = omega  # \omega_0
            self.scale = scale  # \ell^2
            self.name = f"PeriodicTerm_{order}"

        def design_matrix(self) -> JAXArray:
            """The design (also called the feedback) matrix for the PeriodicTerm, $F$"""
            j = self.order
            w = self.omega
            return jnp.array([[0, -w * j], [w * j, 0]])

        def stationary_covariance(self) -> JAXArray:
            """The stationary covariance of the prPeriodicTermocess, Pinf"""
            j = self.order
            arg = 1 / self.scale
            coeff = jax.lax.cond(j == 0, lambda _: 1.0, lambda _: 2.0, j)
            qj2 = coeff * self.Ij(j, arg) / jnp.exp(arg)
            return qj2 * jnp.eye(self.dimension)

        def observation_matrix(self, X: JAXArray) -> JAXArray:
            """The observation matrix for the PeriodicTerm, $H$"""
            return jnp.array([[1, 0]])

        def noise(self) -> JAXArray:
            """The spectral density of the white noise for the PeriodicTerm, $Q_c$"""
            return jnp.zeros((self.dimension, self.dimension))

        def noise_effect_matrix(self) -> JAXArray:
            """The noise effect matrix $L$ for the PeriodicTerm"""
            return jnp.eye(self.dimension)

        def Ij(self, j, x, terms=50) -> JAXArray:
            """
            The modified Bessel function of the first kind, order j, at x.
            Approximated via a truncated Taylor series expansion.
            """
            k = jnp.arange(terms)
            log_terms = (
                -gammaln(k + 1) - gammaln(k + j + 1) + (2 * k + j) * jnp.log(x / 2)
            )
            return jnp.sum(jnp.exp(log_terms))

        def transition_matrix(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
            """The transition matrix A_k for the PeriodicTerm"""
            t1 = self.coord_to_sortable(X1)
            t2 = self.coord_to_sortable(X2)
            dt = t2 - t1
            arg = self.order * self.omega * dt
            cos = jnp.cos(arg)
            sin = jnp.sin(arg)
            return jnp.array([[cos, -sin], [sin, cos]])

        def process_noise(self, X1: JAXArray, X2: JAXArray) -> JAXArray:
            """The process noise Q_k for the PeriodicTerm"""
            del X1, X2
            return jnp.zeros((self.dimension, self.dimension))


class Matern(StateSpaceModel):
    r"""
    A state space implementation of the generic half-integer Matérn kernel
    with smoothness parameter :math:`\nu = p + 1/2` for integer :math:`p \geq 0`,
    length scale :math:`\ell`, and amplitude :math:`\sigma`.

    The model is dimension :math:`d = \nu + 1/2` and has the form 
    (see Eq. 12.34-35 of Särkkä and Solin 2019):

    .. math::

        F = \begin{pmatrix}
                0      & 1 & 0 & & \\
                \vdots & 0 & 1 & & \\
                       &   & \ddots & \ddots & \\
         -a_1\lambda^d & -a_2\lambda^{d-1} & \cdots & -a_d\lambda
            \end{pmatrix}, L = \begin{pmatrix}
                0 \\
                \vdots \\
                0 \\
                1
            \end{pmatrix},
    where :math:`\lambda = \sqrt{2\nu}/\ell` and the coefficients 
    :math:`a_i = \binom{d}{i-1}` are the binomial coefficients.
    The spectral noise density is given by:
        .. math::

            Q_c = \sigma^2 \frac{[(d-1)!]^2}{(2d-2)!} (2\lambda)^{2d-1}.
    
            
    Args:
        nu: The smoothness parameter :math:`\nu` (must be half-integer).
        scale: The parameter :math:`\ell`.
        sigma (optional): The parameter :math:`\sigma`. Defaults to a value of 1.
    """

    nu: JAXArray | float
    scale: JAXArray | float
    sigma: JAXArray | float
    lam: JAXArray | float

    def __init__(
        self,
        nu: JAXArray | float,
        scale: JAXArray | float,
        sigma: JAXArray | float = jnp.ones(()),
        name: str = "Matern",
        **kwargs,
    ):
        assert jnp.isclose(nu % 1, 0.5), (
            "nu must be a half-integer (e.g., 1/2, 3/2, 5/2, etc.)"
        )
        self.nu = nu
        self.scale = scale
        self.sigma = sigma
        self.name = f"Matern{self.nu * 2:.0f}{2}"
        self.lam = jnp.sqrt(2 * self.nu) / self.scale

    @property
    def dimension(self) -> int:
        """The dimension of the state space for the Matérn process"""
        return int(self.nu + 0.5)

    def design_matrix(self) -> JAXArray:
        """The design (also called the feedback) matrix for the Matern process, F"""
        d = self.dimension
        F = jnp.zeros((d, d))
        for i in range(d - 1):
            F = F.at[i, i + 1].set(1)
        for i in range(d):
            ai = factorial(d) / (factorial(i) * factorial(d - i))
            F = F.at[d - 1, i].set(-ai * jnp.power(self.lam, d - i))
        return F

    def observation_matrix(self, X: JAXArray) -> JAXArray:
        """The observation model H for the Matern process"""
        del X
        H = jnp.zeros((1, self.dimension))
        H = H.at[0, 0].set(1)
        return H

    def noise_effect_matrix(self) -> JAXArray:
        """The noise effect matrix L for the Matern process"""
        L = jnp.zeros((self.dimension, 1))
        L = L.at[-1, 0].set(1)
        return L

    def noise(self) -> JAXArray:
        """The scalar Qc for the Matern process"""
        d = self.dimension
        q = (
            jnp.square(self.sigma)
            * jnp.square(factorial(d - 1))
            / factorial(2 * d - 2)
            * jnp.power(2 * self.lam, 2 * d - 1)
        )
        return jnp.array([[q]])

    def stationary_covariance(self) -> JAXArray:
        """The stationary covariance of the Matern-5/2 process, Pinf"""
        from scipy.linalg import solve_continuous_lyapunov

        # TODO: find a JAX version of solve_continuous_lyapunov
        # or figure out the general form for Pinf analytically
        print(
            "Warning: there does not seem to be a JAX implementation "
            "of solve_continuous_lyapunov, so we use the scipy version here "
            "for now. This means that this method will not be JIT-compilable."
        )

        F = self.design_matrix()
        L = self.noise_effect_matrix()
        Qc = self.noise()
        LQL = L @ Qc @ L.T
        return jnp.array(solve_continuous_lyapunov(F, -LQL))



from __future__ import annotations

import jax
import jax.numpy as jnp


def ParallelRTSSmoother(kernel, X, kalman_results):
    """
    Wrapper for Parallel RTS smoother

    Parameters:
        kernel: StateSpaceModel kernel
        X: input coordinates
        kalman_results: output from Kalman filter
            these are the filtered state means (b) and covariances (C)

    Returns:
        E:
        g: smoothed means
        L: smoothed covariances
    """
    mu, P = kalman_results

    Phi = kernel.transition_matrix
    Q = kernel.process_noise
    t = kernel.coord_to_sortable(X)

    asso_params = make_associative_params(Phi, Q, t, mu, P)
    E, g, L = parallel_rts_smoother(asso_params)
    return (E, g, L)


@jax.jit
def make_associative_params(Phi, Q, t, mu, P):
    """Generate the associative parameters needed for parallel RTS

    See eqns in Section 4B of Sarkka & Garcia-Fernandez (2020)
    """

    def make_last_params(mu_last, P_last):
        return (jnp.zeros_like(P_last), mu_last, P_last)

    def make_generic_params(Phi, Q, t_delta, mu, P):
        Phi_dt = Phi(0, t_delta)
        Q_dt = Q(0, t_delta)

        # Placeholder variables (not A, b from parallel KF)
        A = Phi_dt @ P @ Phi_dt.T + Q_dt
        b = P @ Phi_dt.T

        E = jax.scipy.linalg.solve(A.T, b.T, assume_a="pos").T
        g = mu - E @ (Phi_dt @ mu)
        L = P - E @ Phi_dt @ P

        return (E, g, L)

    t_delta = jnp.diff(t)
    E, g, L = jax.vmap(
        make_generic_params,
        in_axes=(None, None, 0, 0, 0),
    )(
        Phi,
        Q,
        t_delta,
        mu[:-1],
        P[:-1],
    )

    EN, gN, LN = make_last_params(mu[-1], P[-1])

    E_all = jnp.concatenate([E, EN[jnp.newaxis, ...]], axis=0)
    g_all = jnp.concatenate([g, gN[jnp.newaxis, ...]], axis=0)
    L_all = jnp.concatenate([L, LN[jnp.newaxis, ...]], axis=0)

    return (E_all, g_all, L_all)


def _combine_per_pair(left, right):
    Ei, gi, Li = left
    Ej, gj, Lj = right

    # The indices need to be swapped for some reason...
    Eij = Ej @ Ei
    gij = Ej @ gi + gj
    Lij = Ej @ Li @ Ej.T + Lj

    return (Eij, gij, Lij)


@jax.jit
def parallel_rts_smoother(asso_params):
    """
    Jax implementation of the parallel RTS smoother algorithm

    See Section 4B of Sarkka & Garcia-Fernandez (2020) for
    a detailed description of the algorithm and notation.

    Total runtime (span) complexity is ~O(logN) where N is the number
    of time steps.
    """
    return jax.lax.associative_scan(
        jax.vmap(_combine_per_pair),
        asso_params,
        reverse=True,
    )



import jax
import jax.numpy as jnp
import tinygp
import smolgp
from smolgp.kernels.base import extract_leaf_kernels

from tests.utils import generate_data
from tests.test_kernels import allclose, offset
from tests.test_kernels import (
    likelihood,
    condition,
    predict,
)

key = jax.random.PRNGKey(0)
jax.config.update("jax_enable_x64", True)


def test_multicomponent():
    # Kernel parameters
    ## SHO Kernel
    w1 = 0.0195
    Q1 = 7.63
    s1 = 0.59
    w2 = 0.0023
    Q2 = 1 / jnp.sqrt(2)
    s2 = 0.329

    ## Matern-5/2 Kernel
    scale = 600.0
    sigma = 5.3

    # State space version
    ssm1 = smolgp.kernels.SHO(omega=w1, quality=Q1, sigma=s1)
    ssm2 = smolgp.kernels.SHO(omega=w2, quality=Q2, sigma=s2)
    ssm3 = smolgp.kernels.Matern52(scale=scale, sigma=sigma)
    ssm_sum = ssm1 + ssm2 + ssm3
    ssm_prod = (ssm1 * ssm2) * ssm3

    # Quasisep version
    qsm1 = tinygp.kernels.quasisep.SHO(omega=w1, quality=Q1, sigma=s1)
    qsm2 = tinygp.kernels.quasisep.SHO(omega=w2, quality=Q2, sigma=s2)
    qsm3 = tinygp.kernels.quasisep.Matern52(scale=scale, sigma=sigma)
    qsm_sum = qsm1 + qsm2 + qsm3
    qsm_prod = (qsm1 * qsm2) * qsm3

    kernels = {"Sum": (ssm_sum, qsm_sum), "Product": (ssm_prod, qsm_prod)}

    for name in kernels:
        ksmol, ktiny = kernels[name]

        ## Generate mock data
        N = 50
        yerr = 0.3
        t_train, y_train = generate_data(N, ktiny, yerr, tmin=0, tmax=1000)
        yerr_train = jnp.full_like(t_train, yerr)

        print(f"Testing {name}...")
        # Build GP objects
        gp_smol = smolgp.GaussianProcess(kernel=ksmol, X=t_train, diag=yerr_train**2)
        gp_tiny = tinygp.GaussianProcess(kernel=ktiny, X=t_train, diag=yerr_train**2)

        # Check likelihood
        likelihood(gp_smol, gp_tiny, y_train, tol=1e-10, atol=1e-11)

        # Check conditioning
        condition(gp_smol, gp_tiny, y_train, tol=1e-9, atol=1e-11)

        # Check predictions
        predict(gp_smol, gp_tiny, y_train, tol=1e-9, atol=1e-11)
        print()

        ## Model decomposition tests for sums of GPs
        if name == "Sum":
            ## Verify component means and predict_component_means agree
            tmin, tmax = t_train.min(), t_train.max()
            dt = 0.1 * (tmax - tmin)
            t_test = jnp.linspace(tmin - dt, tmax + dt, 1000)

            print("  Testing component means...")
            ####################################################
            # smolgp version of component means prediction
            llh, condGP = gp_smol.condition(y_train)
            ## 1) at the data points
            cond_comps = condGP.get_all_component_means(return_var=True)
            ys_ssm = [cond_comps[k][0] for k in cond_comps]
            yvars_ssm = [cond_comps[k][1] for k in cond_comps]
            ## can also get component means at the data points by:
            component_kernels = extract_leaf_kernels(gp_smol.kernel)
            y1, var1 = gp_smol.predict(
                X_test=None, y=y_train, return_var=True, kernel=component_kernels[0]
            )
            allclose(
                "component mean via predict vs get_all_component_means",
                y1 - ys_ssm[0],
                tol=1e-12,
                atol=1e-15,
            )
            allclose(
                "component var via predict vs get_all_component_means",
                var1 - yvars_ssm[0],
                tol=1e-12,
                atol=1e-15,
            )

            ## Can also extract just a component or group of components like so:
            # y_sun, yvar_sun = condGP_ssm.get_component_mean(sunnames, return_var=True)
            ## TODO: add a tinygp version to test this

            ## 2) at predicted points
            predStates = condGP.predict(t_test, return_full_state=True, return_var=True)
            pred_comps = predStates.get_all_components(return_var=True)
            mus_ssm = [pred_comps[k][0] for k in pred_comps]
            vars_ssm = [pred_comps[k][1] for k in pred_comps]
            # Can also extract a group of components' joint prediction
            # mu_sun, var_sun = predStates.get_component(sunnames, return_var=True)
            ## TODO: add a tinygp version to test this

            ####################################################
            # tinygp version of component means prediction
            ## 1) at the data points
            ys_qsm = []
            yvars_qsm = []
            for k in smolgp.kernels.dense.extract_leaf_kernels(gp_tiny.kernel):
                yk, var_k = gp_tiny.predict(y_train, t_train, return_var=True, kernel=k)
                ys_qsm.append(yk)
                yvars_qsm.append(var_k)

            ## 2) at predicted points
            mus_qsm = []
            vars_qsm = []
            for k in smolgp.kernels.dense.extract_leaf_kernels(gp_tiny.kernel):
                yk, var_k = gp_tiny.predict(y_train, t_test, return_var=True, kernel=k)
                mus_qsm.append(yk)
                vars_qsm.append(var_k)

            # Compare
            for i in range(len(mus_ssm)):
                y_res = ys_qsm[i] - ys_ssm[i]
                yvar_res = (yvars_qsm[i] - offset) - yvars_ssm[i]
                allclose(f"component means {i}", y_res, tol=1e-9, atol=1e-12)
                allclose(f"component vars {i}", yvar_res, tol=1e-9, atol=1e-12)

                mu_res = mus_qsm[i] - mus_ssm[i]
                var_res = (vars_qsm[i] - offset) - vars_ssm[i]
                allclose(f"predicted component means {i}", mu_res, tol=1e-9, atol=1e-12)
                allclose(f"predicted component vars {i}", var_res, tol=1e-9, atol=1e-12)

            print()

    # Confirm we don't decomposed beyond Sums
    print("Testing kernel decomposition only separatessums...")
    testkernel = ssm1 + (ssm2 * ssm1)
    components = extract_leaf_kernels(testkernel)
    assert components == [ssm1, (ssm2 * ssm1)], (
        "Kernel was over-decomposed! Recall we cannot decompose products into individual GPs, only sums."
    )
    # print("    Checking names are unique & correct...")
    # # kernel = assign_unique_kernel_names(kernel) # TODO: add this check if it isn't covered naturally by other tests


if __name__ == "__main__":
    test_multicomponent()
    print("All multicomponent kernel tests passed.")



from __future__ import annotations
from collections.abc import Sequence
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    NamedTuple,
)

import equinox as eqx
import jax
import jax.numpy as jnp

from tinygp import kernels, means
from tinygp.helpers import JAXArray
from tinygp.noise import Diagonal, Noise

from smolgp.kernels import StateSpaceModel, Sum, Product
from smolgp.kernels.base import extract_leaf_kernels
from smolgp.kernels.integrated import IntegratedStateSpaceModel

from smolgp.solvers import StateSpaceSolver
from smolgp.solvers import ParallelStateSpaceSolver
from smolgp.solvers.integrated import IntegratedStateSpaceSolver
from smolgp.solvers.integrated import ParallelIntegratedStateSpaceSolver

if TYPE_CHECKING:
    from tinygp.numpyro_support import TinyDistribution

import dataclasses


def assign_unique_kernel_names(kernel: StateSpaceModel) -> StateSpaceModel:
    """
    Return a new kernel where leaf kernel names
    are made unique by appending _1, _2, etc.
    """
    leaves = extract_leaf_kernels(kernel)
    names = [k.name for k in leaves]
    # Early exit if all names are unique (no duplicates)
    if len(set(names)) == len(names):
        return kernel

    # Otherwise, count occurrences
    counts = {}
    for k in leaves:
        counts[k.name] = counts.get(k.name, 0) + 1
    # counter for how many times we've used each duplicated name
    used = {name: 1 for name, c in counts.items() if c > 1}

    def _rename(k: StateSpaceModel) -> StateSpaceModel:
        if isinstance(k, Sum):
            k1 = _rename(k.kernel1)
            k2 = _rename(k.kernel2)
            return Sum(k1, k2)
        if isinstance(k, Product):
            k1 = _rename(k.kernel1)
            k2 = _rename(k.kernel2)
            return Product(k1, k2)
        # Leaf
        if counts[k.name] > 1:
            idx = used[k.name]
            used[k.name] += 1
            newname = f"{k.name}_{idx}"
            return dataclasses.replace(k, name=newname)
        else:
            # Single occurrence: leave unchanged
            return k

    return _rename(kernel)


class ConditionedStates(eqx.Module):
    """
    An object to hold the conditioned means and variances

    X: len(N) data coordinates
    t_states: len(K) time coordinates of all states
    instid  : len(N) instrument ID for each measurement
    obsid   : len(K) observation IDs corresponding to the measurement at each state
    stateid : len(K) state IDs corresponding to each state (0 for exposure-start, 1 for exposure-end)
    predicted_mean/var : len(K) Kalman predicted state
    filtered_mean/var  : len(K) Kalman filtered state
    smoothed_mean/var  : len(K) RTS smoothed state
    """

    X: JAXArray
    t_states: JAXArray
    instid: JAXArray
    obsid: JAXArray
    stateid: JAXArray
    predicted_mean: JAXArray
    filtered_mean: JAXArray
    smoothed_mean: JAXArray
    predicted_cov: JAXArray
    filtered_cov: JAXArray
    smoothed_cov: JAXArray

    def __init__(
        self,
        X,
        t_states: JAXArray,
        instid: JAXArray,
        obsid: JAXArray,
        stateid: JAXArray,
        m_pred: JAXArray,
        P_pred: JAXArray,
        m_filt: JAXArray,
        P_filt: JAXArray,
        m_smooth: JAXArray,
        P_smooth: JAXArray,
    ):
        self.X = X
        self.t_states = t_states
        self.instid = instid
        self.obsid = obsid
        self.stateid = stateid
        self.predicted_mean = m_pred
        self.predicted_cov = P_pred
        self.filtered_mean = m_filt
        self.filtered_cov = P_filt
        self.smoothed_mean = m_smooth
        self.smoothed_cov = P_smooth

    def __call__(self):
        state_coords = (self.t_states, self.instid, self.obsid, self.stateid)
        packaged_results = (
            (self.predicted_mean, self.predicted_cov),
            (self.filtered_mean, self.filtered_cov),
            (self.smoothed_mean, self.smoothed_cov),
        )
        # This should match the output of solver.condition
        return state_coords, packaged_results, None

    def project_at_data(self, observation_model):
        """
        Project the states with measurements (e.g. exposure-ends)
        and sort back into original order as the data
        """

        @jax.jit
        def project(X, m, P):
            H = observation_model(X)
            mu = H @ m
            var = H @ P @ H.T
            return mu, var

        N = jnp.array(self.X).shape[-1]
        ends_idx = jnp.nonzero(self.stateid == 1, size=N)[0]
        sort = jnp.argsort(self.obsid[ends_idx])
        idx = ends_idx[sort]
        m_sel = jnp.take(self.smoothed_mean, idx, axis=0)
        P_sel = jnp.take(self.smoothed_cov, idx, axis=0)
        mu, var = jax.vmap(project)(self.X, m_sel, P_sel)
        return mu.squeeze(), var.squeeze()


class PredictedStates(eqx.Module):
    """
    An object to hold the full predictive states

    t_states: time coordinates at each state
    mean : predictive mean vector for each state
    cov  : predictive covariance for each state
    """

    t_states: JAXArray
    mean: JAXArray
    cov: JAXArray
    kernel: StateSpaceModel

    def __init__(
        self,
        t_states: JAXArray,
        m: JAXArray,
        P: JAXArray,
        kernel: StateSpaceModel,
    ):
        self.t_states = t_states
        self.mean = m
        self.cov = P
        self.kernel = kernel

    def project_mean(self, observation_model) -> JAXArray:
        """
        The projected mean at self.t_states given an observation model.
        """

        def _project(H, m):
            mu = H @ m
            return mu

        H = jax.vmap(observation_model)(self.t_states)
        mu = jax.vmap(_project)(H, self.mean)
        return mu.squeeze()

    def project_variance(self, observation_model) -> JAXArray:
        """
        The projected variance at self.t_states given an observation model.
        """

        def _project(H, P):
            var = H @ P @ H.T
            return var

        H = jax.vmap(observation_model)(self.t_states)
        var = jax.vmap(_project)(H, self.cov)
        return var.squeeze()

    @property
    def loc(self) -> JAXArray:
        """
        The overall mean at the predicted states
        """
        return self.project_mean(self.kernel.observation_matrix)

    @property
    def variance(self) -> JAXArray:
        """
        The overall variance at the predicted states
        """
        return self.project_variance(self.kernel.observation_matrix)

    def get_component(
        self,
        component: str | list[str],
        return_var: bool = False,
    ) -> PredictedStates:
        """
        Extract the predicted states corresponding to a component kernel
        """

        if isinstance(component, str):
            component = [component]

        ## Get effective observation model for the desired component(s)
        def H_comp(X):
            H = self.kernel.observation_model(X, component=component[0])
            for k, name in enumerate(component[1:]):
                H += self.kernel.observation_model(X, component=name)
            return H

        ## Project at test coordinates with component observation model
        component_mean = self.project_mean(H_comp)
        component_var = self.project_variance(H_comp)

        if return_var:
            return component_mean, component_var
        else:
            return component_mean

    def get_all_components(self, return_var: bool = False) -> dict[str, Any]:
        """
        Extract the predicted mean/variance corresponding to each component kernel
        """
        components = extract_leaf_kernels(self.kernel)
        results = {}
        for kernel in components:
            name = kernel.name
            if return_var:
                mu_m, var_m = self.get_component(name, return_var=True)
                results[name] = (mu_m, var_m)
            else:
                mu_m = self.get_component(name, return_var=False)
                results[name] = mu_m
        return results


class GaussianProcess(eqx.Module):
    """An interface for designing a Gaussian Process regression model

    Args:
        kernel (Kernel): The kernel function
        X (JAXArray): The input coordinates. This can be any PyTree that is
            compatible with ``kernel`` where the zeroth dimension is ``N_data``,
            the size of the data set.
        diag (JAXArray, optional): The value to add to the diagonal of the
            covariance matrix, often used to capture measurement uncertainty.
            This should be a scalar or have the shape ``(N_data,)``. If not
            provided, this will default to the square root of machine epsilon
            for the data type being used. This can sometimes be sufficient to
            avoid numerical issues, but if you're getting NaNs, try increasing
            this value.
        noise (Noise, optional): Used to implement more expressive observation
            noise models than those supported by just ``diag``. This can be any
            object that implements the :class:`tinygp.noise.Noise` protocol. If
            this is provided, the ``diag`` parameter will be ignored.
        mean (Callable, optional): A callable or constant mean function that
            will be evaluated with the ``X`` as input: ``mean(X)``
        solver: The solver type to be used to execute the required linear
            algebra.
    """

    num_data: int = eqx.field(static=True)
    dtype: jnp.dtype = eqx.field(static=True)
    kernel: kernels.Kernel
    X: JAXArray
    mean_function: means.MeanBase
    mean: JAXArray
    var: JAXArray | None
    noise: Noise
    solver: StateSpaceSolver
    states: ConditionedStates

    def __init__(
        self,
        kernel: kernels.Kernel,
        X: JAXArray,
        *,
        diag: JAXArray | None = None,
        noise: Noise | None = None,
        mean: means.MeanBase | Callable[[JAXArray], JAXArray] | JAXArray | None = None,
        solver: Any | None = None,
        mean_value: JAXArray | None = None,
        variance_value: JAXArray | None = None,
        covariance_value: Any | None = None,
        states: JAXArray | None = None,
        use_unique_names: bool = True,
        **solver_kwargs: Any,
    ):
        # First, assign unique kernel names if needed
        if use_unique_names:
            self.kernel = assign_unique_kernel_names(kernel)
        else:
            self.kernel = kernel

        # Check if the kernel contains any integrated components
        kernels = extract_leaf_kernels(self.kernel)
        is_integrated = any([isinstance(k, IntegratedStateSpaceModel) for k in kernels])
        is_instantaneous = all([isinstance(k, StateSpaceModel) for k in kernels])

        # If using an integrated solver, ensure X has both coords and bin sizes
        if is_integrated:
            assert isinstance(X, tuple) and len(X) > 1, (
                "IntegratedStateSpaceSolver requires both the data coordinates (e.g. times)"
                " and bin sizes (e.g. exposure times). These should be passed as X=(t, texp)"
                " where t is the midpoint of each measurement and texp is the exposure time"
                " (i.e. each measurement is over the interval [t - texp/2, t + texp/2])."
            )

        # Data coordinates (or tuple of coordinates)
        self.X = X

        # Mean function
        if isinstance(mean, means.MeanBase):
            self.mean_function = mean
        elif mean is None:
            self.mean_function = means.Mean(jnp.zeros(()))
        else:
            self.mean_function = means.Mean(mean)
        if mean_value is None:
            mean_value = jax.vmap(self.mean_function)(self.X)
        self.num_data = mean_value.shape[0]
        self.dtype = mean_value.dtype
        self.mean = mean_value
        self.var = variance_value
        self.states = states
        if self.mean.ndim != 1:
            raise ValueError(
                f"Invalid mean shape: expected ndim = 1, got ndim={self.mean.ndim}"
            )

        # Observation noise model
        if noise is None:
            diag = _default_diag(self.mean) if diag is None else diag
            noise = Diagonal(diag=jnp.broadcast_to(diag, self.mean.shape))
        self.noise = noise

        # Set up the solver
        # TODO: add parallel flag and if so use ParallelIntegratedStateSpaceSolver?
        if solver is None:
            if is_integrated:
                solver = IntegratedStateSpaceSolver
            elif is_instantaneous:
                solver = StateSpaceSolver
            else:
                raise ValueError(
                    "Must provide a solver if the kernel is not "
                    "a StateSpaceModel or IntegratedStateSpaceModel"
                )

            self.solver = solver(
                kernel,
                self.X,
                self.noise,
                **solver_kwargs,
            )
        # If solver type (uninstantiated) is passed
        elif solver in [
            StateSpaceSolver,
            IntegratedStateSpaceSolver,
            ParallelStateSpaceSolver,
            ParallelIntegratedStateSpaceSolver,
        ]:
            self.solver = solver(
                kernel,
                self.X,
                self.noise,
                **solver_kwargs,
            )
        # If a pre-instantiated solver is passed (e.g. like condGP)
        else:
            self.solver = solver

    @property
    def loc(self) -> JAXArray:
        """
        If conditioned, this will be the mean at the data points
        Otherwise, it is just the prior mean.
        """
        return self.mean

    @property
    def variance(self) -> JAXArray:
        """
        If conditioned, this will be the variance at the data points
        Otherwise, it is just the prior variance.
        """
        return self.var

    @property
    def covariance(self) -> JAXArray:
        # TODO: Eq. 12.55 in Sarkka & Solin 2019
        #   if G = states.smoothing_gains exists, otherwise
        #   I guess we raise an error that its not conditioned?
        # return self.covariance_value
        raise NotImplementedError

    def log_probability(self, y: JAXArray) -> JAXArray:
        """Compute the log probability of this multivariate normal

        Args:
            y (JAXArray): The observed data. This should have the shape
                ``(N_data,)``, where ``N_data`` was the zeroth axis of the ``X``
                data provided when instantiating this object.

        Returns:
            The marginal log probability of this multivariate normal model,
            evaluated at ``y``.
        """
        _, _, _, _, v, S = self.solver.Kalman(y, return_v_S=True)
        return self._compute_log_prob(v, S)

    def condition(
        self,
        y: JAXArray,
        X_test: JAXArray | None = None,
        *,
        diag: JAXArray | None = None,  # TODO: is this needed?
        noise: Noise | None = None,  # TODO: is this needed?
        include_mean: bool = True,
        kernel: kernels.Kernel | None = None,  # TODO: select a component kernel
    ) -> ConditionResult:
        """Condition the model on observed data

        Args:
            y (JAXArray): The observed data. This should have the shape
                ``(N_data,)``, where ``N_data`` was the zeroth axis of the ``X``
                data provided when instantiating this object.
            X_test (JAXArray, optional): The coordinates where the prediction
                should be evaluated. This should have a data type compatible
                with the ``X`` data provided when instantiating this object. If
                it is not provided, ``X`` will be used by default, so the
                predictions will be made.
            diag (JAXArray, optional): Will be passed as the diagonal to the
                conditioned ``GaussianProcess`` object, so this can be used to
                introduce, for example, observational noise to predicted data.
            include_mean (bool, optional): If ``True`` (default), the predicted
                values will include the mean function evaluated at ``X_test``.
            kernel (Kernel, optional): A kernel to optionally specify the
                covariance between the observed data and predicted data. See
                :ref:`mixture` for an example.

        Returns:
            A named tuple where the first element ``log_probability`` is the log
            marginal probability of the model, and the second element ``gp`` is
            the :class:`GaussianProcess` object describing the conditional
            distribution evaluated at ``X_test``.
        """

        # If X_test is provided, we need to check that the tree structure
        # matches that of the input data, and that the shapes are all compatible
        # (i.e. the dimension of the inputs must match). This is slightly
        # convoluted since we need to support arbitrary pytrees.
        if X_test is not None:
            matches = jax.tree_util.tree_map(
                lambda a, b: jnp.ndim(a) == jnp.ndim(b)
                and jnp.shape(a)[1:] == jnp.shape(b)[1:],
                self.X,
                X_test,
            )
            if not jax.tree_util.tree_reduce(lambda a, b: a and b, matches):
                raise ValueError(
                    "`X_test` must have the same tree structure as the input `X`, "
                    "and all but the leading dimension must have matching sizes"
                )

        ## Condition on the data and return likelihood ingredients
        conditioned_results = self.solver.condition(y, return_v_S=True)

        ## unpack into prediction at the states
        state_coords, conditioned_states, (v, S) = conditioned_results
        (
            (m_predicted, P_predicted),
            (m_filtered, P_filtered),
            (m_smoothed, P_smoothed),
        ) = conditioned_states

        if isinstance(
            self.solver,
            (IntegratedStateSpaceSolver, ParallelIntegratedStateSpaceSolver),
        ):
            t_states, instid, obsid, stateid = state_coords
        else:
            # If not integrated, t_states = X and id arrays are 'defaulted'
            t_states = self.kernel.coord_to_sortable(state_coords)
            instid = jnp.zeros_like(t_states, dtype=int)
            obsid = jnp.arange(len(t_states), dtype=int)
            stateid = jnp.ones_like(t_states, dtype=int)  # all "have data"

        # Save the conditioned state values to a new GP object
        # so we can use them to make quick predictions at test
        # points with subsequent calls to self.predict
        states = ConditionedStates(
            self.X,
            t_states,
            instid,
            obsid,
            stateid,
            m_predicted,
            P_predicted,
            m_filtered,
            P_filtered,
            m_smoothed,
            P_smoothed,
        )

        ## Grab likelihood (v and S will already be
        ## filtered down to the "at the data" states)
        log_prob = self._compute_log_prob(v, S)

        ## Make predictions at X_test if given
        if kernel is None:
            # If no component kernel passed, use the full model
            observation_model = self.kernel.observation_model
        else:
            # Otherwise use the observation model of the passed
            # kernel, where we zero out all the other components
            observation_model = lambda X: self.kernel.observation_model(
                X, component=kernel.name
            )

        if X_test is not None:
            # If X_test was given, also predit at those points
            mu, var = self.solver.predict(X_test, conditioned_results)
        else:
            # Otherwise, project the conditioned states
            # (at the data points) to observation space
            X_test = self.X
            mu, var = states.project_at_data(observation_model)

        ## Create the conditioned GP
        condGP = GaussianProcess(
            kernel=self.kernel,
            X=X_test,
            noise=self.noise,
            # mean=self.mean,
            solver=self.solver,
            mean_value=mu,
            variance_value=var,
            states=states,
        )

        # Return the likelihood and conditioned GP
        return ConditionResult(log_probability=log_prob, gp=condGP)

    def predict(
        self,
        X_test: JAXArray | None = None,
        y: JAXArray | None = None,
        *,
        return_full_state: bool = False,
        kernel: int | None = None,
        # include_mean: bool = True,
        return_var: bool = False,
        # return_cov: bool = False,
        observation_model: Any | None = None,
    ) -> JAXArray | tuple[JAXArray, JAXArray]:
        """Predict the GP model at new test points conditioned on observed data

        Args:
            X_test (JAXArray, optional): The coordinates where the prediction
                should be evaluated. This should have a data type compatible
                with the ``X`` data provided when instantiating this object. If
                it is not provided, ``X`` will be used by default, so the
                predictions will be made at the data coordinates.
            y (JAXArray): The observed data. Only needs to be given if the GP
                has not yet been conditioned. This should have the shape
                ``(N_data,)``, where ``N_data`` was the zeroth axis of the ``X``
                data provided when instantiating this object.
            include_mean (bool, optional): If ``True`` (default), the predicted
                values will include the mean function evaluated at ``X_test``.
            return_var (bool, optional): If ``True`` (default), the variance of the
                predicted values at ``X_test`` will be returned.
            return_cov (bool, optional): If ``True``, the covariance of the
                predicted values at ``X_test`` will be returned. If
                ``return_var`` is ``True``, this flag will be ignored.
            observation_model (Any, optional): optionally provide a function of
                                X_test to define the output observation model.
                                Default will use that of the kernel.
            return_full_state (bool, optional): If ``True``, return the full predicted state
                mean and covariance, rather than projecting to observation space. Default is
                ``False``, i.e. the result is projected through kernel.observation_model.
            kernel (int, optional): If specified, the index of the kernel in a
                multi-component model (for example, a sum or product of kernels)
                to extract and project (if return_full_state is False) the prediction for.

        Returns:
            The mean of the predictive model evaluated at ``X_test``, with shape
            ``(N_test,)`` where ``N_test`` is the zeroth dimension of
            ``X_test``. If either ``return_var`` or ``return_cov`` is ``True``,
            the variance or covariance of the predicted process will also be
            returned with shape ``(N_test,)`` or ``(N_test, N_test)``
            respectively.
        """

        if self.states is None:
            # Need to condition the GP first
            assert y is not None, (
                "The GP has not been conditioned yet, and no data array `y` was given."
            )
            llh, condGP = self.condition(y)
            return condGP.predict(
                X_test,
                return_full_state=return_full_state,
                kernel=kernel,
                return_var=return_var,
                # return_cov=return_cov,
                observation_model=observation_model,
            )
        else:
            if X_test is None:
                # If no X_test given, predict at the data points
                if return_full_state:
                    mu = self.states.smoothed_mean
                    var = self.states.smoothed_cov
                else:
                    if kernel is None:
                        # already computed here
                        mu, var = self.loc, self.var
                    else:
                        # extract component kernel & project
                        name = kernel if isinstance(kernel, str) else kernel.name
                        H_comp = lambda X: self.kernel.observation_model(
                            X, component=name
                        )
                        mu, var = self.states.project_at_data(H_comp)
            else:
                # Predicting at new test points
                H_test = (
                    self.kernel.observation_model
                    if observation_model is None
                    else observation_model
                )
                mean, variance = self.solver.predict(X_test, self.states())
                if return_full_state:
                    mu = mean
                    var = variance
                    return PredictedStates(
                        t_states=X_test, m=mu, P=var, kernel=self.kernel
                    )
                else:
                    if kernel is not None:
                        name = kernel if isinstance(kernel, str) else kernel.name
                        H_test = lambda X: self.kernel.observation_model(
                            X, component=name
                        )
                    H = jax.vmap(H_test)(X_test)
                    mu = jax.vmap(lambda H_i, m: H_i @ m)(H, mean).squeeze()
                    var = jax.vmap(lambda H_i, P: H_i @ P @ H_i.T)(
                        H, variance
                    ).squeeze()

        if return_var:
            return mu, var
        # if return_cov:
        #     return mu, var
        return mu

    ## TODO: how to define the sample function?
    def sample(
        self,
        key: jax.random.KeyArray,
        shape: Sequence[int] | None = None,
    ) -> JAXArray:
        """Generate samples from the prior process

        Args:
            key: A ``jax`` random number key array. shape (tuple, optional): The
            number and shape of samples to
                generate.

        Returns:
            The sampled realizations from the process with shape ``(N_data,) +
            shape`` where ``N_data`` is the zeroth dimension of the ``X``
            coordinates provided when instantiating this process.
        """
        return self._sample(key, shape)

    def numpyro_dist(self, **kwargs: Any) -> TinyDistribution:
        """Get the numpyro MultivariateNormal distribution for this process"""
        from tinygp.numpyro_support import TinyDistribution

        return TinyDistribution(self, **kwargs)

    @partial(jax.jit, static_argnums=(2,))
    def _sample(
        self,
        key: jax.random.KeyArray,
        shape: Sequence[int] | None,
    ) -> JAXArray:
        raise NotImplementedError
        ## TODO: implement sampling for state space model
        ## fast method to try: https://www.stats.ox.ac.uk/~doucet/doucet_simulationconditionalgaussian.pdf
        ##
        ## or alternatively call the tinygp version? copied below:
        # if shape is None:
        #     shape = (self.num_data,)
        # else:
        #     shape = (self.num_data,) + tuple(shape)
        # normal_samples = jax.random.normal(key, shape=shape, dtype=self.dtype)
        # return self.mean + jnp.moveaxis(
        #     self.solver.dot_triangular(normal_samples), 0, -1
        # )

    @jax.jit
    def _compute_log_prob(self, v: JAXArray, S: JAXArray) -> JAXArray:
        """
        Compute the log-likelihood given v and S from the Kalman filter
        """
        ## More readable version:
        # def llh(k):
        #     v_k, S_k = v[k], S[k]
        #     L_k = jnp.linalg.cholesky(S_k)
        #     w = jax.scipy.linalg.solve_triangular(L_k, v_k, lower=True)
        #     quad = jnp.dot(w, w)
        #     logdetS_k = 2.0 * jnp.sum(jnp.log(jnp.diag(L_k)))
        #     d = v_k.shape[0]
        #     return quad + logdetS_k + d*jnp.log(2*jnp.pi)
        # loglike = -0.5 * jnp.sum(jax.vmap(llh)(jnp.arange(len(v))))

        L = jax.vmap(jnp.linalg.cholesky)(S)  # [T, D, D]
        w = jax.scipy.linalg.solve_triangular(L, v[..., None], lower=True)
        w = jnp.squeeze(w, axis=-1)
        quad = jnp.sum(w**2, axis=1)
        logdetS = 2.0 * jnp.sum(jnp.log(jnp.diagonal(L, axis1=-2, axis2=-1)), axis=1)
        d = v.shape[1]
        log_probs = quad + logdetS + d * jnp.log(2.0 * jnp.pi)
        loglike = -0.5 * jnp.sum(log_probs)

        return jnp.where(jnp.isfinite(loglike), loglike, -jnp.inf)

    def get_component_mean(
        self,
        component: list | str,
        return_var: bool = False,
        **kwargs,
    ) -> Any:
        """
        Get the predictive mean (and variance) of a particular
        (or sum of) component kernel in a multi-component model
        evaluated at self.X

        Args:
            X (JAXArray, optional): The coordinates where the prediction
                should be evaluated. This should have a data type compatible
                with the ``X`` data provided when instantiating this object.
            component (list | str): The name(s) of the component kernel(s)
                to extract the mean for. If a list of names is provided,
                the joint mean and variance for that collection of kernels
                will be returned.
            return_var (bool, optional): If ``True``, also return the variances
                of each component. Default is ``False``.

        Returns:
            If ``return_var`` is ``False``:
                component_mean (JAXArray)
            If ``return_var`` is ``True``:
                component_mean (JAXArray)
                component_var (JAXArray)
        """
        if self.states is None:
            raise ValueError(
                "The GP must be conditioned before getting component means."
            )

        if isinstance(component, str):
            component = [component]

        ## Get effective observation model for the desired component(s)
        def H_comp(X):
            H = self.kernel.observation_model(X, component=component[0])
            for k, name in enumerate(component[1:]):
                H += self.kernel.observation_model(X, component=name)
            return H

        ## Project at data
        component_mean, component_var = self.states.project_at_data(H_comp)

        if return_var:
            return component_mean, component_var
        else:
            return component_mean

    def get_all_component_means(self, return_var: bool = False, **kwargs) -> Any:
        """
        Get the predictive mean (and optionally variance) of each
        component kernels individually, evaluated at self.X

        Args:
            return_var (bool, optional): If ``True``, also return the variances
                of each component. Default is ``False``.

        Returns:
            If ``return_var`` is ``False``, a list of JAX arrays containing the
            means of each component kernel evaluated at the data points.
            If ``return_var`` is ``True``, a tuple where the first element is
            the list of means as before, and the second element is a list of
            JAX arrays containing the variances of each component kernel
            evaluated at the data points.
        """
        if self.states is None:
            raise ValueError(
                "The GP must be conditioned before getting component means."
            )

        ## First, extract all kernels
        kernels = extract_leaf_kernels(self.kernel)

        ## Loop through and project each component
        results = {}
        for k, kernel in enumerate(kernels):
            mu, var = self.get_component_mean(
                component=kernel.name, return_var=True, kwargs=kwargs
            )
            if return_var:
                results[kernel.name] = (mu, var)
            else:
                results[kernel.name] = mu

        return results


class ConditionResult(NamedTuple):
    """The result of conditioning a :class:`GaussianProcess` on data

    This has two entries, ``log_probability`` and ``gp``, that are described
    below.
    """

    log_probability: JAXArray
    """The log probability of the conditioned model

    In other words, this is the marginal likelihood for the kernel parameters,
    given the observed data, or the multivariate normal log probability
    evaluated at the given data.
    """

    gp: GaussianProcess
    """A :class:`GaussianProcess` describing the conditional distribution

    This will have a mean and covariance conditioned on the observed data, but
    it is otherwise a fully functional GP that can sample from or condition
    further (although that's probably not going to be very efficient).
    """


def _default_diag(reference: JAXArray) -> JAXArray:
    """Default to adding some amount of jitter to the diagonal, just in case,
    we use sqrt(eps) for the dtype of the mean function because that seems to
    give sensible results in general.
    """
    return jnp.sqrt(jnp.finfo(reference).eps)



import jax
import jax.numpy as jnp
import smolgp
from jax.scipy.linalg import expm
from tests.utils import allclose

key = jax.random.PRNGKey(0)
jax.config.update("jax_enable_x64", True)


def base(kernel, dts):
    """
    Confirm all analytic versions of matrices
    agree with their numerical versions--
    we're assuming `kernel` has analytic A and Q defined
    """

    # For (underdamped) SHO, these are analytic functions
    transition_matrix_analytic = lambda dt: kernel.transition_matrix(0, dt)
    process_noise_analytic = lambda dt: kernel.process_noise(0, dt)

    # Numerical versions
    # transition_matrix_numerical = lambda dt: expm(F * dt)
    transition_matrix_numerical = lambda dt: super(
        type(kernel), kernel
    ).transition_matrix(0, dt)
    process_noise_vanloan = lambda dt: super(type(kernel), kernel).process_noise(
        0, dt, use_van_loan=True
    )
    process_noise_fromPinf = lambda dt: super(type(kernel), kernel).process_noise(
        0, dt, use_van_loan=False
    )

    # Test all dts
    print("Comparing analytic Phi vs. expm(F*dt)...")
    A_analytic = jax.vmap(transition_matrix_analytic)(dts)
    A_numerical = jax.vmap(transition_matrix_numerical)(dts)
    allclose(
        "Transition matrix",
        A_analytic - A_numerical,
        tol=1e-9,
        atol=1e-12,
    )
    print("Comparing analytic Q vs. VanLoan Q...")
    Q_analytic = jax.vmap(process_noise_analytic)(dts)
    Q_vanloan = jax.vmap(process_noise_vanloan)(dts)
    allclose(
        "Process noise (VanLoan)",
        Q_analytic - Q_vanloan,
        tol=1e-9,
        atol=1e-12,
    )
    print("Comparing analytic Q vs. Q from Pinf...")
    Q_fromPinf = jax.vmap(process_noise_fromPinf)(dts)
    allclose(
        "Process noise (from Pinf)",
        Q_analytic - Q_fromPinf,
        tol=1e-9,
        atol=1e-12,
    )
    print("All base matrix tests passed.")


def integrated_transition(kernel, dts):
    """
    Confirm all analytic versions of matrices and their
    augmented forms agree with their numerical versions--
    we're assuming `kernel` has analytic Phibar defined
    """

    ### Integrated transition matrix
    Phibar_analytic = lambda dt: kernel.integrated_transition_matrix(0, dt)
    Phibar_from_VanLoan = lambda dt: super(
        type(kernel), kernel
    ).integrated_transition_matrix(0, dt)
    print("Comparing analytic Phibar vs. Van Loan...")
    A_analytic = jax.vmap(Phibar_analytic)(dts)
    A_numerical = jax.vmap(Phibar_from_VanLoan)(dts)
    allclose(
        "Integrated transition matrix",
        A_analytic - A_numerical,
        tol=1e-8,
        atol=1e-12,
    )

    ### Augmented transition matrix
    F_aug = kernel.design_matrix()
    Phiaug_numerical = lambda dt: expm(F_aug * dt)
    Phiaug_analytic = lambda dt: kernel.transition_matrix(0, dt)
    print("Comparing analytic Phiaug vs. expm(Faug*dt)...")
    Aaug_analytic = jax.vmap(Phiaug_analytic)(dts)
    Aaug_numerical = jax.vmap(Phiaug_numerical)(dts)
    allclose(
        "Augmented transition matrix",
        Aaug_analytic - Aaug_numerical,
        tol=1e-5,
        atol=1e-12,
    )


def integrated_process_noise(kernel, dts, tol=1e-8):
    # Augmented process noise
    Qaug_implemented = lambda dt: kernel.process_noise(0, dt)
    Qaug_from_VanLoan = lambda dt: kernel.process_noise(0, dt, force_numerical=True)
    print("Comparing implemented Qaug vs. VanLoan...")
    Qaug_analytic = jax.vmap(Qaug_implemented)(dts)
    Qaug_vanloan = jax.vmap(Qaug_from_VanLoan)(dts)
    allclose(
        "Augmented process noise",
        Qaug_analytic - Qaug_vanloan,
        tol=tol,
        atol=1e-12,
    )
    print("All integrated/augmented matrix tests passed.")


def test_proofs():
    ## TODO: do we also want to test shapes here?

    ## This test is done with the SHO Kernel, since
    ## we have analytic versions to compare against
    S = 2.36
    w = 0.0195
    Q = 7.63
    sigma = jnp.sqrt(S * w * Q)
    sho = smolgp.kernels.SHO(omega=w, quality=Q, sigma=sigma)
    isho = smolgp.kernels.IntegratedSHO(omega=w, quality=Q, sigma=sigma)

    ## Test a wide dynamic range of Deltas
    ## NOTE: expm breaks down around ~1e6
    dts = jnp.logspace(-6, 5, 1000)

    print("Testing base (instantaneous) kernel matrices...")
    base(sho, dts)
    print()
    print("Testing integrated kernel matrices...")
    integrated_transition(isho, dts)

    ## Separately test small Delta and large Delta for process noise
    dtsmall = dts[dts < 1e3]
    dtlarge = dts[dts >= 1e3]
    integrated_process_noise(isho, dtsmall, tol=1e-6)
    integrated_process_noise(isho, dtlarge, tol=1e-3)

    ## and verify dt=0 is correct
    assert jnp.all(isho.process_noise(0.0, 0.0) == 0), (
        "Process noise at dt=0 should be zero."
    )
    assert jnp.all(isho.transition_matrix(0.0, 0.0) == jnp.eye(isho.dimension)), (
        "Transition matrix at dt=0 should be the identity."
    )


if __name__ == "__main__":
    test_proofs()
    print("All proof tests passed.")



import numpy as np
import jax.numpy as jnp
import matplotlib.pyplot as plt

## Colors for benchmarking plots
colors = {"SSM" : "#1f77b4",
          "QSM" : "#ff7f0e", 
          "GP"  : "#2ca02c", 
          "pSSM": "#6A0E95"
          }
markers = {"SSM" : "o",
           "QSM" : "s", 
           "GP"  : "D", 
           "pSSM": "*"
          }
markersize = {"SSM" : 8,
              "QSM" : 6, 
              "GP"  : 6, 
              "pSSM": 10,
          }

def symlogticks(vmin, vmax, linthresh=1e-16, spacing=1):
    """Generate ticks for a symlog axis."""
    thresh = int(jnp.log10(linthresh))
    negticks = -10.**np.arange(thresh, vmin, spacing)[::-1]
    posticks = 10.**jnp.arange(thresh, vmax, spacing)
    return jnp.concatenate([negticks, posticks])

def scale_nans(runtime_array, Ns, power=1):
    # Find the last non-nan index
    last_valid_idx = jnp.where(~jnp.isnan(runtime_array[:, 0]))[0][-1]
    last_valid_runtime = runtime_array[last_valid_idx, 0]
    last_valid_N = Ns[last_valid_idx]

    # Scale the nans based on the last valid point
    for i in range(last_valid_idx + 1, len(runtime_array)):
        runtime_array = runtime_array.at[i, 0].set(
            last_valid_runtime * (Ns[i] / last_valid_N) ** power
        )
        runtime_array = runtime_array.at[i, 1].set(0)  # Set std to 0 for scaled points

    return runtime_array

def plot_benchmark(
    Ns,
    runtimes,
    ax=None,
    savefig=None,
    scale=True,
    powers={"SSM": 1, "QSM": 1, "GP": 3, "pSSM": 1},
    labels=None,
):
    """
    runtimes should be a dict with values (mean, std) for each N
    """
    if labels is None:
        labels = {name: name for name in runtimes}

    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(6, 6), sharex=True)

    for name in runtimes:
        runtime_array = jnp.array(runtimes[name])
        mean_runtime = runtime_array[:, 0]
        std_runtime = runtime_array[:, 1]
        if scale:
            scaled = scale_nans(runtime_array, Ns, power=powers[name])
            ax.errorbar(Ns, scaled[:, 0], scaled[:, 1], c=colors[name], ls=":")
        ax.errorbar(Ns, mean_runtime, std_runtime, c=colors[name], marker=markers[name],
                     markersize=markersize[name], fmt='-', label=labels[name])

    ax.legend()
    ax.set(
        xscale="log", yscale="log", xlabel="Number of data points", ylabel="Runtime [s]"
    )
    tens = 10 ** jnp.arange(jnp.log10(Ns[0]), jnp.log10(Ns[-1]) + 1)
    ax.set_xticks(tens, labels=["" for _ in tens], minor=True)
    ax.grid(alpha=0.5, zorder=-10)
    ax.grid(alpha=0.5, zorder=-10, which="minor", axis="x")
    # ax.set_ylim(top=jnp.nanmax(runtime_ss[:,0])*1e3)
    
    if savefig:
        plt.savefig(savefig, dpi=300, bbox_inches="tight")
    return ax


"""
``smolgp`` is designed to be a drop-in extension of the `tinygp <https://github.com/dfm/tinygp>`_
library for building Gaussian Process (GP) models in Python. As such, it is also built on top of
`jax <https://github.com/google/jax>`_. The driving design philosophy is to match the API of ``tinygp``
as closely as possible. With only a few exceptions, any existing code you have that uses ``tinygp`` should
work with ``smolgp`` by simply by finding-and-replacing ``tiny`` with ``smol``.


``smolgp`` uses the state space
representations of Gaussian Processes to implement linear-time (or up to logN with
parallelization on GPU) solvers for GP regression and forecasting. It also implements
"integrated" kernels that can model time-averaged measurements, such as those from
long-exposure astronomical observations, which can also be solved in linear time and are
also compatible with the parallel methods.

The primary way that you interact with ``smolgp`` is to construct
"kernel" functions using the building blocks provided in the ``kernels``
subpackage (see :ref:`api-kernels`), and then passing that to a
:class:`GaussianProcess` object to do all the computations. Check out the
:ref:`tutorials` for a more complete introduction.
"""

from smolgp import (
    kernels as kernels,
    solvers as solvers,
)
from smolgp.gp import GaussianProcess as GaussianProcess
# from smolgp.smolgp_version import __version__ as __version__



from __future__ import annotations
from typing import Any
import jax
import jax.numpy as jnp
import equinox as eqx

from tinygp.helpers import JAXArray
from tinygp.noise import Noise
from tinygp.solvers.quasisep.solver import QuasisepSolver
from smolgp.kernels.base import StateSpaceModel
from smolgp.solvers.integrated.parallel.kalman import ParallelIntegratedKalmanFilter
from smolgp.solvers.integrated.parallel.rts import ParallelIntegratedRTSSmoother


class ParallelIntegratedStateSpaceSolver(eqx.Module):
    """
    A solver that uses ``jax.lax.associative_scan`` to implement
    parallel Kalman filtering and RTS smoothing for integrated measurements
    """

    X: JAXArray
    kernel: StateSpaceModel
    noise: Noise
    state_coords: JAXArray
    _state_coords: JAXArray

    def __init__(
        self,
        kernel: StateSpaceModel,
        X: JAXArray,
        noise: Noise,
    ):
        """Build a :class:`IntegratedStateSpaceSolver` for a given kernel and coordinates

        Args:
            kernel: The kernel function.
            X: The input coordinates. The coordinates for an integrated model should be a tuple of
                    X = (t, delta, instid),
                where `t` is the usual coordinate (e.g. time) at the measurements (midpoints),
                `delta` is the integration range (e.g. exposure time) for each measurement,
                and `instid` is an index encoding which instrument the measurement corresponds to.
            noise: The noise model for the process.
            state_coords: Bookkeeping indices for the discretized states used in Kalman/RTS
        """
        self.kernel = kernel
        self.X = X
        self.noise = noise

        ## Preprocess state coordinates (exposure start/stops)
        ## and assign labels to each observation/state for bookkeeping:
        ## obsid   -- array len(K): which observation (0,...,N-1) is being made at each state k
        ## instids -- array len(N): which instrument (0,...,Ninst-1) recorded observation n
        ## stateid -- array len(K): 0 for exposure-start, 1 for exposure-end
        tmid, delta, instid = self.X  # unpack coordinates

        ## Construct interleaved time array of chronological exposure start/stop times
        ts = tmid - delta / 2  # Exposure start times
        te = tmid + delta / 2  # Exposure end times
        obsid = jnp.arange(len(tmid)).repeat(2)

        # Interleave start and end times into one array (fastest)
        # https://stackoverflow.com/questions/5347065/interleaving-two-numpy-arrays-efficiently
        t_states = jnp.empty((ts.size + te.size,), dtype=tmid.dtype)
        t_states = t_states.at[0::2].set(ts)  # evens are start times
        t_states = t_states.at[1::2].set(te)  # odds are end times
        stateid = jnp.tile(jnp.array([0, 1]), len(tmid))  # 0 for start, 1 for end
        # Have to re-sort because exposures can overlap
        # enforce end times before start times at same t
        sortidx = jnp.lexsort((-stateid, t_states))
        t_states = t_states[sortidx]
        obsid = obsid[sortidx]
        stateid = stateid[sortidx]  # 0 for t_s, 1 for t_e

        # Pack-up state_coords for Kalman and RTS functions
        _instid = jnp.repeat(instid, repeats=2)[sortidx]
        self.state_coords = (t_states, instid, obsid, stateid)
        self._state_coords = (t_states, _instid, obsid, stateid)

    def normalization(self) -> JAXArray:
        # TODO: do we want/can we implement this in state space? for now, fall back to quasisep
        return QuasisepSolver(self.kernel, self.X, self.noise).normalization()

    def Kalman(self, y, return_v_S=True) -> Any:
        """Wrapper for Kalman filter used with this solver"""
        t_states, instid, obsid, stateid = self._state_coords
        return ParallelIntegratedKalmanFilter(
            self.kernel,
            self.X,
            y,
            t_states,
            obsid,
            instid,
            stateid,
            self.noise,
            return_v_S=return_v_S,
        )

    def RTS(self, kalman_results) -> Any:
        """Wrapper for RTS smoother used with this solver"""
        t_states, instid, obsid, stateid = self._state_coords
        return ParallelIntegratedRTSSmoother(
            self.kernel,
            t_states,
            stateid,
            instid,
            kalman_results,
        )

    def condition(self, y, return_v_S=True) -> JAXArray:
        """
        Compute the Kalman predicted, filtered, and RTS smoothed
        means and covariances at each of the input coordinates
        """

        # Kalman filtering
        kalman_results = self.Kalman(y, return_v_S=return_v_S)
        if return_v_S:
            m_filtered, P_filtered, m_predicted, P_predicted, v, S = kalman_results
            v_S = (v, S)
        else:
            m_filtered, P_filtered, m_predicted, P_predicted = kalman_results
            v_S = None

        # RTS smoothing
        rts_results = self.RTS((m_predicted, P_predicted, m_filtered, P_filtered))
        m_smoothed, P_smoothed = rts_results

        conditioned_states = (
            (m_predicted, P_predicted),
            (m_filtered, P_filtered),
            (m_smoothed, P_smoothed),
        )

        return self.state_coords, conditioned_states, v_S

    @jax.jit
    def predict(self, X_test, conditioned_results) -> JAXArray:
        """
        Algorithm for making predictions at arbitrary coordinates X_test

        Args:
            X_test              : The test coordinates.
            conditioned_results : The output of self.condition()
            observation_model   : (optional) H for the test points
                                  should be a function just like
                                  self.kernel.observation_model

        There are three cases:
            1. Retrodiction  : smoothing from the first data point
                               using the prior as the prediction
            2. Interpolation : filtering from most recent data point
                               and smoothing from next future point
            3. Extrapolation : predicting from final filtered point
        """

        # Unpack conditioned results
        state_coords, conditioned_states, _ = conditioned_results
        (
            (m_predicted, P_predicted),
            (m_filtered, P_filtered),
            (m_smoothed, P_smoothed),
        ) = conditioned_states
        t_states, instid, obsid, stateid = state_coords

        # Unpack test coordinates
        t_test = self.kernel.coord_to_sortable(X_test)

        # Array shapes
        # N = len(self.X)  # number of data points (unused)
        K = len(t_states)  # number of states
        M = len(t_test)  # number of test points

        # Prior covariance for retrodiction
        Pinf = self.kernel.stationary_covariance()
        if not isinstance(Pinf, JAXArray):  # if multicomponent model
            Pinf = Pinf.to_dense()  # needs to be array form here

        # Prior mean for retrodiction
        # mean = jnp.zeros(self.kernel.d)  # TODO: mean function of base kernel
        # m0 = jnp.block([mean] + self.kernel.num_insts * [jnp.zeros(self.kernel.d)])
        m0 = jnp.zeros(self.kernel.dimension)

        # Nearest/next past/future state for each datapoint
        k_nexts = jnp.searchsorted(t_states, t_test, side="right")

        # Method to use for test point
        past = k_nexts <= 0  # Retrodict
        future = k_nexts >= K  # Extrapolate
        during = ~past & ~future  # Interpolate
        cases = past.astype(int) * 0 + during.astype(int) * 1 + future.astype(int) * 2

        # Shorthand for matrices
        A_aug = lambda dt: self.kernel.transition_matrix(0, dt)
        Q_aug = lambda dt: self.kernel.process_noise(0, dt)

        def kalman(k_prev, ktest):
            """
            Kalman prediction from most recent
            filtered (but not RTS smoothed) state
            """
            dt = t_test[ktest] - t_states[k_prev]
            m_k = m_filtered[k_prev]
            P_k = P_filtered[k_prev]
            A_star = A_aug(dt)
            Q_star = Q_aug(dt)
            m_star_pred = A_star @ m_k
            P_star_pred = A_star @ P_k @ A_star.T + Q_star
            return m_star_pred, P_star_pred

        def smooth(k_next, ktest, m_star_pred, P_star_pred):
            """
            RTS smooth the prediction (ktest) using
            the nearest future data point (k_next)

            m_star_pred and P_star_pred are the output of kalman(k, k_star)
            """
            # Next (future) predicted & smoothed state
            m_pred_next = m_predicted[k_next]
            P_pred_next = P_predicted[k_next]
            m_hat_next = m_smoothed[k_next]
            P_hat_next = P_smoothed[k_next]

            # Transition matrix
            dt = t_states[k_next] - t_test[ktest]
            A_k = A_aug(dt)

            # RTS update
            G_k = jnp.linalg.solve(P_pred_next.T, (P_star_pred @ A_k.T).T).T
            m_star_hat = m_star_pred + G_k @ (m_hat_next - m_pred_next)
            P_star_hat = P_star_pred + G_k @ (P_hat_next - P_pred_next) @ G_k.T

            return m_star_hat, P_star_hat

        def retrodict(ktest):
            """Reverse-extrapolate from first datapoint t_star"""
            m_star, P_star = smooth(0, ktest, m0, Pinf)
            return m_star, P_star

        def interpolate(ktest):
            """Interpolate between nearest data points"""

            # Get nearest data point before and after the test point
            k_next = k_nexts[ktest]
            k_prev = k_next - 1

            # 1. Kalman predict from most recent data point (in past)
            m_star_pred, P_star_pred = kalman(k_prev, ktest)

            # 2. RTS smooth from next nearest data point (in future)
            m_star_hat, P_star_hat = smooth(k_next, ktest, m_star_pred, P_star_pred)

            # return project(ktest, m_star_hat, P_star_hat)
            return m_star_hat, P_star_hat

        def extrapolate(ktest):
            """Kalman predict from from last datapoint t_star"""
            m_star, P_star = kalman(-1, ktest)
            return m_star, P_star

        @jax.jit
        def predict_point(ktest):
            """
            Switch between retrodiction, interpolation, and extrapolation
            for a single test point ktest
            """
            return jax.lax.switch(
                cases[ktest], (retrodict, interpolate, extrapolate), (ktest)
            )

        # Calculate predictions
        ktests = jnp.arange(0, M, 1)
        (pred_mean, pred_var) = jax.vmap(predict_point)(ktests)

        return pred_mean, pred_var
